#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2009, 2011-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=W0201


from glob import glob
import os
import sys
import sysconfig
from contextlib import suppress
from collections import OrderedDict

use_setuptools = not os.getenv('PYBIK_NO_SETUPTOOLS')
if use_setuptools:
    try:
        import setuptools
    except ImportError:
        use_setuptools = False
if not use_setuptools:
    from distutils.core import setup
    from distutils.cmd import Command
    from distutils.extension import Extension
    from distutils import dist
    from distutils.command.install import install as install_orig
else:
    print('using setuptools')
    from setuptools import setup
    from setuptools import Command
    from setuptools import Extension
    from setuptools import dist
    from setuptools.command.install import install as install_orig

from distutils.dir_util import remove_tree
from distutils.file_util import move_file
from distutils.dep_util import newer, newer_group
from distutils.log import warn, info, error, debug

import distutils.command.build
import distutils.command.build_ext
import distutils.command.build_scripts
import distutils.command.install_data
import distutils.command.install_lib
import distutils.command.clean
import distutils.command.sdist

#from Cython.Distutils import build_ext

#print('using', 'setuptools' if use_setuptools else 'distutils')

# the clean command should leave a clean tree
sys.dont_write_bytecode = True
os.environ['PYTHONDONTWRITEBYTECODE'] = '1'

import buildlib.utils
import pybiklib
from pybiklib import config

os.environ['QT_SELECT'] = '5'


class Distribution (distutils.dist.Distribution):
    def __init__(self, attrs=None):
        self.bug_contact = None
        self.appmeta_templates = None
        #XXX: Mangle alpha/beta/rc versions to PEP440 version scheme for setuptools, pip, pypi ...
        version = attrs['version']
        pep440 = version.replace('~alpha', 'a').replace('~beta', 'b').replace('~rc', 'rc')
        attrs['version'] = pep440
        distutils.dist.Distribution.__init__(self, attrs=attrs)
        if version != pep440:
            self.metadata.version = version


class build(distutils.command.build.build):
    """Adds extra commands to the build target."""
    user_options = distutils.command.build.build.user_options + [
            ('inplace', 'i', 'ignore build-lib and put compiled modules into the source directory'),
            #XXX: In Python 3.5 distutils.command.build.build has a parallel option
            ('parallel=', None, 'number of parallel build jobs or "True"'),
            ('arch-only', None, 'Build only architecture dependent files'),
            ('indep-only', None, 'Build only architecture independent files'),
            ('pickle-protocol=', None, 'Pickle protocol for modeldata (default: highest protocol)'),
            ('maxsize=', None, 'Maximum model size (default: 10)'),
            ('reproducible', None, 'Build modeldata reproducible (experimental)'),
            ('variants=',    None, 'Build extension variants'),
            ('armhf',        None, 'Cross build for armhf'),
            ('debugflags=',  None, 'Comma-separated list of debug flags'),
        ]
    boolean_options = distutils.command.build.build.boolean_options + [
            'inplace', 'arch-only', 'indep-only', 'reproducible',
            'armhf',
        ]
        
    def has_pure_modules(self):
        return self.indep_only and distutils.command.build.build.has_pure_modules(self)
        
    _commands_dict = dict(distutils.command.build.build.sub_commands)
    # pylint: disable=W0212
    sub_commands = [('build_py',      has_pure_modules),
                    ('build_ext',     lambda self: self.arch_only and self._commands_dict['build_ext'](self)),
                    ('build_scripts', lambda self: self.indep_only and self._commands_dict['build_scripts'](self)),
                    ('build_models',  lambda self: self.indep_only),
                    ('build_appicons',lambda self: self.indep_only and not self.inplace),
                    ('build_i18n',    lambda self: self.indep_only),
                    ('build_appmeta', lambda self: self.indep_only),
                    ('build_man',     lambda self: self.indep_only and not self.inplace),
                ]
    # pylint: enable=W0212
                
    def initialize_options(self):
        distutils.command.build.build.initialize_options(self)
        self.inplace = 0
        self.parallel = None
        self.arch_only = False
        self.indep_only = False
        self.pickle_protocol = '-1'
        self.maxsize = None
        self.reproducible = False
        self.variants = None
        self.armhf = False
        self.debugflags = None
        
    def finalize_options(self):
        # In Python 3.5 distutils.command.build.build has a parallel option
        # convert 'true' to True before the base class fails
        if isinstance(self.parallel, str) and self.parallel.lower() == 'true':
            self.parallel = True
        distutils.command.build.build.finalize_options(self)
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
        
        
class build_csrc (distutils.command.build_ext.build_ext):
    description = 'build C source code with py2pyx and cython'
    accepted_variants = buildlib.utils.accepted_variants()
    user_options = distutils.command.build_ext.build_ext.user_options + [
        ('cython-opts=', None,          'Cython options'),
        ('variants=',    None,          'Build extension variants (comma-separated list of: %s)' %
                                        buildlib.utils.variant_string()),
        ('armhf',        None,          'Cross build for armhf'),
        ('debugflags=',  None,          'Comma-separated list of debug flags'),
    ]
    
    def initialize_options(self):
        distutils.command.build_ext.build_ext.initialize_options(self)
        self.inplace = None
        self.cython_opts = ''
        self.variants = None
        self.armhf = None
        self.debugflags = None
        self.cc_files = []
        
    def finalize_options(self):
        # deactivate parallel build in Python 3.5, does not work with this command
        self.parallel = 1
        distutils.command.build_ext.build_ext.finalize_options(self)
        self.set_undefined_options('build',
                                       ('inplace', 'inplace'),
                                       ('variants', 'variants'),
                                       ('armhf', 'armhf'),
                                       ('debugflags', 'debugflags'),
                                    )
        if self.variants is None:
            self.variants = []
        else:
            self.variants = self.variants.split(',')
            for v in self.variants:
                if v not in self.accepted_variants:
                    raise ValueError('unknown variant: ' + v)
                    
        if self.armhf:
            from distutils.util import get_platform
            self.build_lib = self.build_lib.replace(get_platform(), 'linux-armv7l')
            self.build_temp = self.build_temp.replace(get_platform(), 'linux-armv7l')
        self.cython_opts = self.cython_opts.split()
        import pybiklib.debug
        if self.debugflags is not None:
            pybiklib.debug.set_flagstring(self.debugflags)
        
        # aliases
        groups = []
        aliases = OrderedDict({'': ''})
        for v in buildlib.utils.variants:
            if v['type'] is None or v['group'] in groups:
                continue
            groups.append(v['group'])
            if v['type'] == 'enum':
                for a, va in list(aliases.items()):
                    aliases[a] = va + (v['id'] if v['group'] != 'glvariant' else '???')
                    for ve in buildlib.utils.variants:
                        if ve['type'] == 'enum' and ve['group'] == v['group']:
                            aliases[a+ve['id']] = va+ve['id']
            elif v['type'] == 'bool':
                for a, va in list(aliases.items()):
                    aliases[a+v['id']] = va+v['id']
        self.aliases = OrderedDict((k,v) for k,v in aliases.items() if k != v)
        
    def make_files(self, infiles, outfiles, func, args, exec_msg=None, skip_msg=None, level=1):
        """Works like make_file, but for operations
        that generate one or more output files.
        """
        if skip_msg is None:
            skip_msg = "skipping %s (inputs unchanged)" % outfiles
        if isinstance(infiles, str):
            infiles = (infiles,)
        elif not isinstance(infiles, (list, tuple)):
            raise TypeError("'infiles' must be a string, or a list or tuple of strings")
        if exec_msg is None:
            exec_msg = "generating %s from %s" % (outfiles, ', '.join(infiles))
        if self.force or any(newer_group(infiles, outfile) for outfile in outfiles):
            self.execute(func, args, exec_msg, level)
        else:
            debug(skip_msg)
            
    def compare_file(self, filename1, filename2):
        if self.force or not os.path.exists(filename2):
            return False
        with open(filename1, 'rb') as file1, open(filename2, 'rb') as file2:
            return file1.read() == file2.read()
            
    @staticmethod
    def run_py2pyx(py_file, pyx_file, pxd_file, pxm_file, pxc_file):
        info('py2pyx: %s --> %s' % (py_file, pyx_file))
        from buildlib.py2pyx import create_pyx, Py2pyxError
        try:
            create_pyx(py_file, pyx_file, pxd_file, pxm_file, pxc_file)
        except Py2pyxError as e:
            error("error: %s", e)
            sys.exit(1)
        
    identifier_translate = bytes((i if bytes([i]).isalnum() else ord('_')) for i in range(256))
    
    @classmethod
    def replace_cwd(cls, filename):
        with open(filename, 'rb') as f:
            text = f.read()
        cwdb_esc = bytes(ord(c) for c in os.getcwd()[:32] if ord(c)<128).translate(cls.identifier_translate)
        text = text.replace(os.getcwdb(), b'/tmp').replace(cwdb_esc, b'_tmp')
        with open(filename, 'wb') as f:
            f.write(text)
            
    def run_cython(self, cython_opts, infile, outfile):
        from Cython.Compiler.CmdLine import parse_command_line
        from Cython.Compiler.Main import compile
        
        info('cython: %s --> %s' % (infile, outfile))
        args = ['-3', '-o', outfile] + cython_opts + [infile]
        options, unused_sources = parse_command_line(args)
        try:
            result = compile([infile], options)
        except Exception as e:
            error("error: %s", e)
            sys.exit(1)
        if result.num_errors > 0:
            sys.exit(1)
        self.replace_cwd(outfile)
        
    def build_pyx_(self, py_file, pyx_file, pxd_file, pxm_file, pxc_file):
        pxd_file_tmp = pxd_file + '.tmp'
        pxm_file_tmp = pxm_file + '.tmp'
        pxc_file_tmp = pxc_file + '.tmp'
        self.run_py2pyx(py_file, pyx_file, pxd_file_tmp, pxm_file_tmp, pxc_file_tmp)
        for px_file, px_file_tmp in ((pxd_file, pxd_file_tmp), (pxm_file, pxm_file_tmp), (pxc_file, pxc_file_tmp)):
            if os.path.exists(px_file_tmp):
                if self.compare_file(px_file_tmp, px_file):
                    os.remove(px_file_tmp)
                    info("unchanged file '%s'", px_file)
                else:
                    if os.path.exists(px_file):
                        os.remove(px_file)
                    move_file(px_file_tmp, px_file)
            elif os.path.exists(px_file):
                os.remove(px_file)
                
    def build_pyx(self, py_file, pyx_file, pxd_file, pxm_file, pxc_file):
        self.mkpath(os.path.dirname(pyx_file))
        self.make_files([py_file], pyx_file, self.build_pyx_, (py_file, pyx_file, pxd_file, pxm_file, pxc_file))
        
    @staticmethod
    def find_qt_command(command):
        from shutil import which
        return which(command) or which(command+'-qt5') or command
        
    def create_extension_variant(self, extension, variantflags, variant_map):
        variantstr = ''.join(variantflags)
        info('generating extension %r with variant %r', extension.name, variantstr)
        def filename_copy(filename):
            return os.path.join(self.build_temp, filename)
        def make_copy(in_file, out_file):
            self.mkpath(os.path.dirname(out_file))
            self.copy_file(in_file, out_file)
        def filename_variant(base, ext):
            if '_' in base.lstrip('_'):
                src_base, src_variant = base.rsplit('_', 1)
                if src_variant in self.accepted_variants:
                    if src_variant not in variantflags:
                        return None
                    base = src_base
            return os.path.join(self.build_temp, base+'_'+variantstr+ext)
        def make_variant(in_file, out_file):
            self.mkpath(os.path.dirname(out_file))
            self.copy_file(in_file, out_file)
            from buildlib.utils import modify_file
            info('changing to variant %r in %r, %s', variantstr, out_file, variantflags)
            for modulename, mvariantflags in variant_map.items():
                mvariantstr = ''.join(mvariantflags)
                modified = modify_file(out_file, [(r'\[\[%s_VARIANT\]\]'%modulename, modulename+'_'+mvariantstr)], count=0)
                if not modified:
                    warn('unchanged file: %r for module-variant: %s_%s', out_file, modulename, mvariantstr)
            for v in variantflags:
                vgroup = self.accepted_variants[v]['group'].upper()
                vid = self.accepted_variants[v]['id']
                subs = [(r'\[\[%s\]\]' % vgroup, vid)]
                if self.accepted_variants[v]['type'] == 'bool':
                    subs.append((r'\b(DEF %s = )False' % vgroup, r'\1True'))
                modified = modify_file(out_file, subs, count=0)
                if not modified:
                    warn('unchanged file: %r for variant %r', out_file, v)
            modified = modify_file(out_file, [(r'\[\[VARIANT\]\]', variantstr)], count=0)
            if not modified:
                warn('unchanged file: %r for variant %r', out_file, variantstr)
            def depends_lines():
                yield '{'
                for vs, dvs in self.variant_map.items():
                    joinkey = repr(''.join(vs)) + ':'
                    joinvalue = list(k + '_' + ''.join(vs) for k, vs in dvs.items())
                    yield '  {:12} {!r},'.format(joinkey, joinvalue)
                yield '}'
            modified = modify_file(out_file, [(r'\[\[DEPENDS\]\]', '\n  '.join(depends_lines()))])
            def aliases_lines():
                yield '{'
                for avs, vs in self.aliases.items():
                    yield '  {:12} {!r},'.format(repr(avs) + ':', vs)
                yield '}'
            modified = modify_file(out_file, [(r'\[\[ALIASES\]\]', '\n  '.join(aliases_lines()))])
            
        def get_libs():
            for template in extension.libraries:
                vlibs = variant_libraries.get(template)
                if vlibs is None:
                    yield template
                else:
                    for v in variantflags:
                        yield from vlibs.get(v, [])
        extvariant = Extension(extension.name+'_'+variantstr, sources=[],
                               depends=[],
                               language=extension.language,
                               define_macros=extension.define_macros,
                               include_dirs=extension.include_dirs,
                               libraries=list(get_libs()),
                              )
        for in_file in extension.sources:
            base, ext = os.path.splitext(in_file)
            out_file = filename_variant(base, ext)
            if out_file is None:
                continue
            self.make_file([in_file], out_file, make_variant, [in_file, out_file])
            extvariant.sources.append(out_file)
        for in_file in extension.depends:
            base, ext = os.path.splitext(in_file)
            if ext == '.py':
                continue
            if 'qtq' not in variantflags and os.path.basename(in_file) == 'qtq.pxd':
                continue
            if os.path.isfile(base+'.pyx'):
                out_file = filename_variant(base, ext)
                if out_file is None:
                    continue
                self.make_file([in_file], out_file, make_variant, [in_file, out_file])
            else:
                out_file = filename_copy(in_file)
                self.make_file([in_file], out_file, make_copy, [in_file, out_file])
            extvariant.depends.append(out_file)
        return extvariant
        
    def build_extensions(self):
        def gen_extensions(variants, variantflags, selected=True):
            if variants:
                group, vs, all_vs = variants[0]
                for v in all_vs:
                    yield from gen_extensions(variants[1:],
                                              variantflags if v is None else variantflags + ((group, v),),
                                              selected and v in vs)
            else:
                yield selected, variantflags
                
        variants = list(buildlib.utils.group_variants(self.variants))
        self.variant_map = OrderedDict()
        for extension in self.extensions:
            built_variants = []
            modulename = extension.name.split('/')[-1]
            for selected, variantflags in gen_extensions(variants, ()):
                filtered_variantflags = tuple(vf for g, vf in variantflags
                                                 if g in variantgroups_by_module[modulename])
                variantflags = tuple(vf for g, vf in variantflags)
                if 'd' not in variantflags and modulename == '_debug':
                    continue
                if 'qtw' not in variantflags and modulename == '_qtui':
                    continue
                variant_map = self.variant_map.setdefault(variantflags, OrderedDict())
                assert modulename not in variant_map
                if modulename != '_qtexec':
                    variant_map[modulename] = filtered_variantflags
                if not selected:
                    continue
                if filtered_variantflags in built_variants:
                    continue
                extvariant = self.create_extension_variant(extension, filtered_variantflags, variant_map)
                self.build_extension(extvariant)
                built_variants.append(filtered_variantflags)
                        
    def build_extension(self, extension):
        sources = extension.sources
        depends = sources + extension.depends
            
        py_files = []
        pyx_files = []
        pyx_files_dep = []
        ui_files = []
        moc_files = []
        c_files = []
        c_files_dep = []
        
        for in_file in extension.sources:
            base, ext = os.path.splitext(in_file)
            if ext == '.py':
                py_files.append(in_file)
            elif ext == '.pyx':
                pyx_files.append(in_file)
            else:
                assert False, in_file
        assert len(py_files) <= 1
        for in_file in extension.depends:
            base, ext = os.path.splitext(in_file)
            if ext == '.py':
                basename = os.path.basename(base)
                dirname = os.path.dirname(base)
                pxd_file = os.path.join(dirname, '_%s.pxd' % basename)
                pyx_files_dep.append(pxd_file)
            elif ext == '.pxd':
                pyx_files_dep.append(in_file)
            elif ext == '.ui':
                ui_files.append(in_file)
            elif ext == '.h':
                c_files_dep.append(in_file)
            else:
                assert False, in_file
            
        # generate files from py-files with #px annotations
        for in_file in py_files:
            base, ext = os.path.splitext(in_file)
            assert ext == '.py'
            basename = os.path.basename(base)
            dirname = os.path.dirname(base)
            pyx_file = os.path.join(dirname, '_%s.pyx' % basename)
            pxd_file = os.path.join(dirname, '_%s.pxd' % basename)
            pxm_file = os.path.join(dirname, '_%s_moc.h' % basename)
            pxc_file = os.path.join(dirname, '_%s_priv.cpp' % basename)
            self.build_pyx(in_file, pyx_file, pxd_file, pxm_file, pxc_file)
            pyx_files.append(pyx_file)
            if os.path.exists(pxd_file):
                pyx_files_dep.append(pxd_file)
            if os.path.exists(pxm_file):
                c_files_dep.append(pxm_file)
                moc_files.append(pxm_file)
            if os.path.exists(pxc_file):
                c_files.append(pxc_file)
                
        def use_cc(out_file):
            inplace_file = os.path.join('csrc', os.path.basename(out_file))
            if os.path.exists(inplace_file) and not self.force:
                self.mkpath(os.path.dirname(out_file))
                self.copy_file(inplace_file, out_file)
                return True
            return False
            
        # generate C-files with cython
        if extension.language == 'c++':
            c_file_ext = '.cpp'
            cython_opts = ['--cplus'] + self.cython_opts
        else:
            c_file_ext = '.c'
            cython_opts = self.cython_opts
        assert len(pyx_files) <= 1, pyx_files
        for in_file in pyx_files:
            base, ext = os.path.splitext(in_file)
            assert ext == '.pyx'
            out_file = base + c_file_ext
            if not use_cc(out_file):
                self.make_file(pyx_files+pyx_files_dep, out_file,
                               self.run_cython, (cython_opts, in_file, out_file))
            c_files.append(out_file)
            
        # generate C++-files with uic
        uic_cmd = self.find_qt_command('uic')
        for in_file in ui_files:
            base, ext = os.path.splitext(in_file)
            assert ext == '.ui'
            out_file = base + '.h'
            if not use_cc(out_file):
                self.make_file([in_file], out_file, self.spawn, (
                        [uic_cmd, in_file, '-o', out_file, '--translate', 'gettext_translate'],))
            c_files_dep.append(out_file)
            
        # generate C++-files with moc
        moc_cmd = self.find_qt_command('moc')
        for in_file in moc_files:
            base, ext = os.path.splitext(in_file)
            assert in_file.endswith('_moc.h')
            out_file = base + '.cpp'
            if not use_cc(out_file):
                self.make_file([in_file], out_file, self.spawn, ([moc_cmd, in_file, '-o', out_file],))
            c_files.append(out_file)
            
        extension.depends = c_files_dep
        extension.sources = c_files
        self.cc_files += c_files + c_files_dep
        
        
class build_ext (build_csrc):
    description = distutils.command.build_ext.build_ext.description
    
    def get_ext_filename(self, ext_name):
        filename = distutils.command.build_ext.build_ext.get_ext_filename(self, ext_name)
        if self.armhf:
            from distutils.sysconfig import get_config_var
            multiarch = get_config_var('MULTIARCH')
            filename = filename.replace(multiarch, 'arm-linux-gnueabihf')
        return filename
        
    def build_extension(self, extension):
        if sys.flags.debug:
            extension.define_macros.extend([('QT_DECLARATIVE_DEBUG', None), ('QT_QML_DEBUG', None)])
        if self.armhf:
            multiarch = 'arm-linux-gnueabihf'
        else:
            multiarch = sysconfig.get_config_var('MULTIARCH')
            if multiarch is None:
                multiarch = sysconfig.get_config_var('multiarchsubdir') or ''
                multiarch = multiarch.strip('/')
        extension.include_dirs = [d.replace('<multiarchsubdir>', multiarch)
                                  for d in extension.include_dirs]
        sources = extension.sources
        ext_path = self.get_ext_fullpath(extension.name)
        depends = sources + extension.depends
        if self.force or newer_group(depends, ext_path, 'newer'):
            info('building C-Code for %r extension', extension.name)
        else:
            debug('skipping %r extension (up-to-date)', extension.name)
            return
            
        # build C-code
        build_csrc.build_extension(self, extension)
        # build extension module from C-code
        if self.armhf:
            info('cross building for armhf')
            os.environ['CC'] = 'arm-linux-gnueabihf-gcc'
            os.environ['CXX'] = 'arm-linux-gnueabihf-g++'
        with suppress(AttributeError, ValueError):
            import distutils.sysconfig
            distutils.sysconfig.customize_compiler(self.compiler)
            if extension.language == 'c++':
                #XXX: distutils uses the wrong compiler (not a problem with gcc) and flags (-Wstrict-prototypes) for C++ code
                #     silence cc1plus: warning: command line option '-Wstrict-prototypes' is valid for C/ObjC but not for C++
                #     http://bugs.python.org/issue1222585
                #     http://bugs.python.org/issue9031
                if '-Wstrict-prototypes' in self.compiler.compiler_so:
                    info("removing command line option '-Wstrict-prototypes': option is valid for C/ObjC but not for C++")
                    self.compiler.compiler_so.remove('-Wstrict-prototypes')
                if self.compiler.compiler_cxx[0] != self.compiler.compiler[0] == self.compiler.compiler_so[0]:
                    info('replacing compiler {} by {}'.format(self.compiler.compiler_so[0], self.compiler.compiler_cxx[0]))
                    self.compiler.compiler_so[0] = self.compiler.compiler_cxx[0]
        if self.armhf:
            self.compiler.linker_so[0] = self.compiler.compiler_so[0]
        distutils.command.build_ext.build_ext.build_extension(self, extension)
        #HACK: Due to build_csrc.compare_file the C compiler may not run, even though
        #      the dependencies expect this. Therefore update the timestamp manually.
        ext_path = self.get_ext_fullpath(extension.name)
        try:
            os.utime(ext_path, None)
        except OSError:
            pass
        
        
class build_scripts (distutils.command.build_scripts.build_scripts):
    user_options = distutils.command.build_scripts.build_scripts.user_options + [
            ('inplace',     'i', 'ignore build-lib and put compiled modules into the source directory'),
            ('build-temp=', 't', "temporary build directory"),
        ]
    boolean_options = distutils.command.build.build.boolean_options + ['inplace']
    
    def initialize_options(self):
        distutils.command.build_scripts.build_scripts.initialize_options(self)
        self.inplace = None
        self.build_temp = None
        
    def finalize_options(self):
        distutils.command.build_scripts.build_scripts.finalize_options(self)
        self.set_undefined_options('build',
                    ('inplace', 'inplace'),
                    ('build_temp', 'build_temp'),
                )
                
    def run(self):
        build_dir = self.build_dir
        self.build_dir = self.build_temp
        distutils.command.build_scripts.build_scripts.run(self)
        self.build_dir = build_dir
        for script in self.scripts:
            outfile = os.path.basename(script)
            script = os.path.join(self.build_temp, outfile)
            if not os.path.exists(script):
                continue
            if not self.inplace:
                self.mkpath(self.build_dir)
                outfile = os.path.join(self.build_dir, outfile)
            outfile, ext = os.path.splitext(outfile)
            if ext != '.py':
                outfile += ext
            self.copy_file(script, outfile)
        
        
class build_models(Command):
    description = 'build data for Pybik models'
    user_options = [
        ('inplace',     'i', 'ignore build-lib and put compiled UI modules into the source '
                             'directory alongside your pure Python modules'),
        ('force',       'f', 'forcibly build everything (ignore file timestamps)'),
        ('parallel=',   None,'number of parallel build jobs or "True"'),
        ('pickle-protocol=', None, 'Pickle protocol for modeldata (default: highest protocol)'),
        ('maxsize=',    None, 'Maximum model size (default: 10)'),
        ('reproducible', None, 'Build modeldata reproducible (experimental)'),
        ('debugflags=', None, 'Comma-separated list of debug flags'),
    ]
    boolean_options = ['inplace', 'force', 'reproducible']
    
    def initialize_options(self):
        self.build_base = None
        self.inplace = None
        self.force = None
        self.parallel = None
        self.pickle_protocol = None
        self.maxsize = None
        self.reproducible = None
        self.debugflags = None
        
    def finalize_options(self):
        self.set_undefined_options('build',
                                       ('build_base', 'build_base'),
                                       ('inplace', 'inplace'),
                                       ('force', 'force'),
                                       ('parallel', 'parallel'),
                                       ('pickle_protocol', 'pickle_protocol'),
                                       ('maxsize', 'maxsize'),
                                       ('reproducible', 'reproducible'),
                                   )
        if isinstance(self.parallel, str):
            if self.parallel.lower() == 'true':
                self.parallel = True
            else:
                self.parallel = int(self.parallel)
        self.pickle_protocol = -1 if self.pickle_protocol is None else int(self.pickle_protocol)
        self.reproducible = False if self.reproducible is None else bool(self.reproducible)
        if self.debugflags is not None:
            import pybiklib.debug
            pybiklib.debug.set_flagstring(self.debugflags)
                
    def run(self):
        from buildlib import modeldata
        if self.maxsize is not None:
            from ast import literal_eval
            maxsize = literal_eval(self.maxsize)
            if isinstance(maxsize, tuple):
                modeldata.minsize, modeldata.maxsize = maxsize
            else:
                modeldata.maxsize = maxsize
        modeldir = 'data/models' if self.inplace else os.path.join(self.build_base, 'share', 'models')
        self.mkpath(modeldir)
        import pybiklib.debug
        prefix = 'f' if pybiklib.debug.DEBUG_MODELFAST else 'd'
        if self.force and not self.dry_run:
            for filename in os.listdir(modeldir):
                if filename[0] == prefix:
                    filename = os.path.join(modeldir, filename)
                    os.remove(filename)
        if self.force:
            testfunc = None
        else:
            testfunc = lambda filename: (any(newer(os.path.join('buildlib', f), filename) for f in
                                                ('modeldata.py', 'modeldef.py', 'geom.py')))
        if testfunc is None or testfunc(modeldata.get_indexfilename(modeldir)):
            message = 'generating model data, this may take some time'
            def create_modeldata():
                modeldata.create_modeldata(modeldir, testfunc, self.parallel,
                                        self.pickle_protocol, self.reproducible)
            self.execute(create_modeldata, [], msg=message)
        data_files = self.distribution.data_files
        if self.dry_run:
            return
        for filename in os.listdir(modeldir):
            sourcepath = os.path.join(modeldir, filename)
            data_files.append(('share/pybik/models', (sourcepath,)))
        
        
class build_appicons(Command):
    description = 'rename Pybik app-icons'
    user_options = [
        ('force',       'f', 'forcibly build everything (ignore file timestamps)'),
    ]
    boolean_options = ['inplace', 'force', 'reproducible']
    
    def initialize_options(self):
        self.build_base = None
        
    def finalize_options(self):
        self.set_undefined_options('build',
                                       ('build_base', 'build_base'),
                                   )
                
    def run(self):
        if self.dry_run:
            return
            
        sourcedir = 'data/icons'
        data_files = self.distribution.data_files
        for filename in os.listdir(sourcedir):
            size = str(int(filename.rsplit('-', 1)[1].split('.')[0]))
            sourcepath = os.path.join(sourcedir, filename)
            builddir = os.path.join(self.build_base, 'share', 'icons', size)
            self.mkpath(builddir)
            buildpath = os.path.join(builddir, 'pybik.png')
            self.copy_file(sourcepath, buildpath)
            data_files.append(('share/icons/hicolor/{0}x{0}/apps'.format(size), [buildpath]))
            
            
class build_i18n(Command):
    description = 'compile message catalog to binary format'
    
    user_options = [
        ('inplace',         'i',  'ignore build-lib and put compiled UI modules into the source '
                                  'directory alongside your pure Python modules'),
        ('force',           'f',  'forcibly build everything (ignore file timestamps)'),
    ]
        
    boolean_options = ['inplace', 'force']
    
    def initialize_options(self):
        self.build_base = None
        self.inplace = None
        self.force = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('inplace', 'inplace'),
                                            ('force', 'force'),
                                  )
            
    def run(self):
        """Compile message catalog to binary format"""
        if self.inplace:
            mo_dir = 'data/locale'
        else:
            mo_dir = os.path.join(self.build_base, 'share', 'locale')
        data_files = self.distribution.data_files
        domain = self.distribution.metadata.name
        
        for lang in buildlib.utils.get_linguas():
            po_file = os.path.join('po', lang+'.po')
            mo_dir_lang =  os.path.join(mo_dir, lang, "LC_MESSAGES")
            mo_file = os.path.join(mo_dir_lang, "%s.mo" % domain)
            self.mkpath(mo_dir_lang)
            def msgfmt(po_file, mo_file):
                self.spawn(["msgfmt", po_file, "-o", mo_file])
            self.make_file([po_file], mo_file, msgfmt, [po_file, mo_file])
            
            targetpath = os.path.join("share/locale", lang, "LC_MESSAGES")
            data_files.append((targetpath, (mo_file,)))
            
            
class build_appmeta(Command):
    description = 'build translated desktop files'
    
    user_options = [
        ('inplace',  'i',  'ignore build-lib and put compiled UI modules into the source directory'),
        ('force',    'f',  'forcibly build everything (ignore file timestamps)'),
    ]
        
    boolean_options = ['inplace', 'force']
    
    def initialize_options(self):
        self.build_base = None
        self.inplace = None
        self.force = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('inplace', 'inplace'),
                                            ('force', 'force'),
                                  )
        self.appmeta_dir = 'data/app-meta' if self.inplace else os.path.join(self.build_base, 'share',' app-meta')
            
    def run(self):
        from buildlib import create_docs
        
        if self.distribution.appmeta_templates is None:
            return
            
        data_files = self.distribution.data_files
        self.mkpath(self.appmeta_dir)
        
        def make_translated_file(src, tmp, dst, msgfmt_type):
            create_docs.create_doc(src, tmp, skip='deb')
            self.spawn(["msgfmt", msgfmt_type, '-d', 'po', '--template', tmp, '-o', dst])
            os.remove(tmp)
        for targetpath, template, template_depends in self.distribution.appmeta_templates:
            templatefile = create_docs.get_template_filename(template)
            tmpfile = os.path.join(self.appmeta_dir, os.path.basename(templatefile))
            dstfile = os.path.join(self.appmeta_dir, os.path.basename(template))
            if dstfile.endswith('.desktop'):
                msgfmt_type = '--desktop'
            elif dstfile.endswith('.appdata.xml'):
                msgfmt_type = '--xml'
            else:
                self.warn('Skipping %s, seems not to be a supported template.' % templatefile)
                continue
            self.make_file(glob('po/*.po') + [templatefile] + template_depends, dstfile,
                           make_translated_file, [template, tmpfile, dstfile, msgfmt_type])
            data_files.append((targetpath, [dstfile]))
        
        
class build_man (Command):
    description = 'build the manpage'
    user_options = [
        ('force',       'f',  'forcibly build everything (ignore file timestamps)'),    ]
    
    def initialize_options(self):
        self.force = None
        self.build_base = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('force', 'force'),                                  )
        
    def run(self):
        from buildlib import create_docs
        
        def create_manpage(man_template, man_rst, man_file):
            man_template = os.path.splitext(man_template)[0]
            create_docs.create_doc(man_template, man_rst)
            self.spawn(['rst2man', man_rst, man_file])
            
        man_dir = os.path.join(self.build_base, 'share', 'man')
        self.mkpath(man_dir)
        tmp_dir = os.path.join(self.build_base, 'share', 'temp', 'rst')
        self.mkpath(tmp_dir)
        if self.dry_run:
            return
        for man_template in glob('data/templates/manpage_*.rst.in'):
            man_rst, unused_ext = os.path.splitext(os.path.basename(man_template))
            base, unused_ext = os.path.splitext(man_rst)
            unused, section, script = base.split('_', 2)
            man_rst = os.path.join(tmp_dir, man_rst)
            man_file = os.path.join(man_dir, '.'.join((script, section)))
            config.MANPAGE_SECTION = section
            self.make_file(['pybiklib/config.py', man_template], man_file,
                           create_manpage, [man_template, man_rst, man_file])
            self.distribution.data_files.append(('share/man/man'+section, [man_file]))
            
            
class build_doc (Command):
    description = "build documentation files"
    user_options = [
        ('inplace',         'i',  'ignore build-lib and put compiled UI modules into the source '
                                  'directory alongside your pure Python modules'),
    ]
    boolean_options = ['inplace']
    
    def initialize_options(self):
        self.inplace = None
        self.build_base = None
        self.doc_files = []
        
    def finalize_options(self):
        self.set_undefined_options('build', ('inplace', 'inplace'), ('build_base', 'build_base'))
        
    def run(self):
        from buildlib import create_docs
        
        info('generating documentation files')
        doc_dir = os.path.join(self.build_base, 'share', 'doc')
        self.mkpath(doc_dir)
        for basename, skip in [
                    ('copyright', 'deb'),
                    ('README', None),
                    ('INSTALL', None),
                ]:
            filename = None if self.inplace else os.path.join(doc_dir, basename)
            info('generating file: %r', filename or basename)
            create_docs.create_doc(basename, filename, skip=skip)
            if filename:
                self.doc_files.append(filename)
        
        
class install_lib (distutils.command.install_lib.install_lib):
    user_options = distutils.command.install_lib.install_lib.user_options + [
            ('arch-only', None, 'Install only architecture dependent files'),
            ('indep-only', None, 'Install only architecture independent files'),
        ]
    boolean_options = ['arch-only', 'indep-only']
    
    def initialize_options(self):
        distutils.command.install_lib.install_lib.initialize_options(self)
        self.data_dir = None
        self.arch_only = None
        self.indep_only = None
        
    def finalize_options(self):
        distutils.command.install_lib.install_lib.finalize_options(self)
        self.set_undefined_options('install',
                                       ('arch_only', 'arch_only'),
                                       ('indep_only', 'indep_only'),
                                   )
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
            
    def build(self):
        if not self.skip_build:
            if self.distribution.has_pure_modules() and self.indep_only:
                self.run_command('build_py')
            if self.distribution.has_ext_modules() and self.arch_only:
                self.run_command('build_ext')
                
                
class install_desktop(Command):
    description = 'install desktop files'
    user_options = [('ask', None, 'Ask before installation'),]
    boolean_options = ['ask']
    
    def initialize_options(self):
        self.ask = False
        
    def finalize_options(self):
        pass
    
    def run(self):
        def ask():
            try:
                answer = input('\nInstall desktop file? [Yn]: ')
            except EOFError:
                print()
                return False
            if answer.lower() not in ['y', '']:
                return False
            return True
                
        self.run_command('build_appmeta')
        appmeta_dir = self.get_finalized_command('build_appmeta').appmeta_dir
        target_dir = config.get_data_home('applications')
        
        def keyabspath(line, key, file=None):
            key += '='
            if line.startswith(key):
                if file is None:
                    file = line.split('=', 1)[1].rstrip()
                line = key + os.path.abspath(file) + '\n'
            return line
            
        self.mkpath(target_dir)
        for fnin in glob('%s/*.desktop' % appmeta_dir):
            fnout = os.path.join(target_dir, os.path.basename(fnin))
            tout = ''
            with open(fnin, 'rt', encoding='utf-8') as fin:
                for line in fin:
                    line = keyabspath(line, 'TryExec')
                    line = keyabspath(line, 'Exec')
                    line = keyabspath(line, 'Icon', config.APPICON192_FILE)
                    tout += line
            with suppress(FileNotFoundError), open(fnout, 'rt', encoding='utf-8') as fout:
                if tout == fout.read():
                    info('desktop file %r unchanged', fnout)
                    return
            if self.ask and not ask():
                return
            info('installing desktop file to: %s', fnout)
            with open(fnout, 'wt', encoding='utf-8') as fout:
                fout.write(tout)
                    
                    
class install (install_orig):
    user_options = install_orig.user_options + [
            ('data-dir=', 't', 'Directory where the application will find the data'),
            ('arch-only', None, 'Install only architecture dependent files'),
            ('indep-only', None, 'Install only architecture independent files'),
        ]
    boolean_options = ['arch-only', 'indep-only']
        
    def initialize_options(self):
        install_orig.initialize_options(self)
        self.data_dir = None
        self.arch_only = False
        self.indep_only = False
        
    def finalize_options(self):
        install_orig.finalize_options(self)
        if self.data_dir is None:
            self.data_dir = os.path.join(self.install_data, 'share')
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
        if not self.indep_only:
            self.__class__.sub_commands = [(cmd, func)
                            for cmd, func in install_orig.sub_commands
                                if cmd == 'install_lib']
            
    def run(self):
        if not self.skip_build:
            # install_orig.run() will run build, but we need
            # to modify a file between build and install
            build_cmd = self.distribution.get_command_obj('build')
            build_cmd.arch_only = self.arch_only
            build_cmd.indep_only = self.indep_only
            self.run_command('build')
            self.skip_build = True
            
        filename = os.path.join(self.build_lib, 'pybiklib', 'config.py')
        app_data_dir = os.path.join(self.data_dir, 'pybik')
        if self.indep_only:
            from buildlib.utils import modify_file
            modified = modify_file(filename, [
                    (r'^(data_dir\s*=\s*).*$',      r'\1' + repr(self.data_dir)),
                    (r'^(appdata_dir\s*=\s*).*$',   r'\1' + repr(app_data_dir)),])
            if not modified:
                warn('unchanged file: %r', filename)
        install_orig.run(self)
        
        
class clean (distutils.command.clean.clean):
    user_options = distutils.command.clean.clean.user_options + [
            ('inplace', 'i',         'clean up files in the source directory'),
        ]
    boolean_options = distutils.command.clean.clean.boolean_options + [
            'inplace',
        ]
        
    def initialize_options(self):
        distutils.command.clean.clean.initialize_options(self)
        self.inplace = None
        
    def finalize_options(self):
        distutils.command.clean.clean.finalize_options(self)
        self.set_undefined_options('build',
                                       ('inplace', 'inplace'),
                                   )
        
    def run(self):
        def remove_tree_(directory):
            if os.path.exists(directory):
                remove_tree(directory, dry_run=self.dry_run, verbose=self.verbose)
            else:
                warn("%r does not exist -- can't clean it", directory)
        def remove_file(filename):
            if os.path.exists(filename):
                info('removing %r', filename)
                if not self.dry_run:
                    os.remove(filename)
            else:
                warn("%r does not exist -- can't clean it", filename)
                
        if self.all:
            remove_tree_(os.path.join(self.build_base, 'share'))
        distutils.command.clean.clean.run(self)
        remove_file('MANIFEST')
        if self.inplace:
            for dirname in ('buildlib/', 'pybiklib/', 'pybiklib/ext/', 'pybiklib/plugins/', 'tools/'):
                remove_tree_(dirname + '__pycache__')
            if self.all:
                dirnames = ['data/app-meta', 'data/locale', 'data/models']
                for dirname in dirnames:
                    remove_tree_(dirname)
                for filename in glob('pybiklib/ext/*.so'):
                    remove_file(filename)
                for filename in ['copyright', 'INSTALL', 'README', 'pybik']:
                    remove_file(filename)
                
                
class sdist(distutils.command.sdist.sdist):
    user_options = distutils.command.sdist.sdist.user_options + [
            ('debian-names', None,  'Create archive files with Debian names'),
        ]
    boolean_options = distutils.command.sdist.sdist.boolean_options + [
            'debian-names'
        ]
        
    def initialize_options(self):
        distutils.command.sdist.sdist.initialize_options(self)
        self.debian_names = False
        self.appmeta_dir = None
        self.distribution.get_option_dict('build_csrc')['variants'] = ('class sdist', 'ogl,es2')
        
    def finalize_options(self):
        distutils.command.sdist.sdist.finalize_options(self)
        self.set_undefined_options('build_appmeta', ('appmeta_dir', 'appmeta_dir'))
        
    def run(self):
        self.run_command('build_doc')
        self.run_command('build_csrc')
        self.run_command('build_appmeta')
        distutils.command.sdist.sdist.run(self)
        
        for archive_file in self.archive_files:
            if self.debian_names:
                debian_file = archive_file.replace('-', '_', 1).replace('.tar', '.orig.tar', 1)
                os.rename(archive_file, debian_file)
        
    def get_file_list(self):
        for lang in buildlib.utils.get_linguas():
            self.filelist.files.append(os.path.join('po', lang+'.po'))
        distutils.command.sdist.sdist.get_file_list(self)
        
    def make_release_tree(self, base_dir, files):
        distutils.command.sdist.sdist.make_release_tree(self, base_dir, files)
        # doc files
        for filename in self.get_finalized_command('build_doc').doc_files:
            self.copy_file(filename, base_dir)
        # csrc
        dst_dir = os.path.join(base_dir, 'csrc')
        self.mkpath(dst_dir)
        cc_files = self.get_finalized_command('build_csrc').cc_files
        for src in cc_files:
            self.copy_file(src, dst_dir)
        # desktop- and appdata-files
        if self.distribution.appmeta_templates is not None:
            dst_dir = os.path.join(base_dir, 'data/app-meta')
            self.mkpath(dst_dir)
            for unused_targetpath, template, unused_template_depends in self.distribution.appmeta_templates:
                src = os.path.join(self.appmeta_dir, template)
                self.copy_file(src, dst_dir)
        #
        dst_dir = os.path.join(base_dir, 'po')
        self.mkpath(dst_dir)
        self.copy_file('po/LINGUAS', dst_dir)
        
        
variantgroups_by_module = {
                        '_debug': ('glvariant',),
                        '_gldraw': ('glvariant', 'gldebug'),
                        '_glarea': ('glvariant', 'gldebug', 'offscreen'),
                        '_qtui': (),
                        '_qt': ('qtvariant', 'glvariant', 'gldebug', 'offscreen'),
                        '_qtexec': (),
                     }
                    
scripts = ['pybiklib/pybik.py']
data_files = [
                ('share/pybik/ui/cursors', glob('data/ui/cursors/*')),
                ('share/pybik/ui/images', glob('data/ui/images/*')),
                ('share/pybik/ui/qt', glob('data/ui/qt/*.*')),
                ('share/pybik/ui/qt/images', glob('data/ui/qt/images/*')),
                ('share/pybik/ui/thumbnails', glob('data/ui/thumbnails/*')),
                ('share/pybik/plugins', glob('data/plugins/*')),
                ('share/pybik/shaders', glob('data/shaders/*')),
                ('share/pybik/', ['data/GPL-3']),
            ]
appmeta_templates = [ ('share/applications', 'pybik.desktop', []),
                      ('share/metainfo', 'net.launchpad.pybik.appdata.xml', ['NEWS']),
                    ]
variant_libraries = {
                'GL': {'ogl': ['GL'],
                       'es2': ['GLESv2']},
                'Qt': {'qtw': ['Qt5Core', 'Qt5Gui', 'Qt5Widgets'],
                       'qtq': ['Qt5Core', 'Qt5Gui', 'Qt5Widgets', 'Qt5Qml', 'Qt5Quick']},
            }
ext_modules = [
                Extension('pybiklib/ext/_debug',
                            glob('pybiklib/ext/_debug_*.pyx'),
                            depends=glob('pybiklib/ext/gl_*.pxd')+glob('pybiklib/ext/_debug_*.pxd'),
                            libraries=['GL'],
                            define_macros=[('GL_GLEXT_PROTOTYPES', None)]),
                Extension('pybiklib/ext/_gldraw',
                            ['pybiklib/ext/gldraw.py'],
                            depends=glob('pybiklib/ext/gl_*.pxd'),
                            libraries=['GL'],
                            define_macros=[('GL_GLEXT_PROTOTYPES', None)]),
                Extension('pybiklib/ext/_glarea',
                            ["pybiklib/ext/glarea.py"],
                            depends=glob('pybiklib/ext/gl_*.pxd')+['pybiklib/ext/gldraw.py'],
                            libraries=['GL'],
                            define_macros=[('GL_GLEXT_PROTOTYPES', None)]),
                Extension('pybiklib/ext/_qtui',
                            ["pybiklib/ext/qtui.py"],
                            depends=['pybiklib/ext/qt.pxd', 'pybiklib/ext/qtui_p.h'] + glob('data/ui/qt/*.ui'),
                            language='c++',
                            include_dirs=[sysconfig.get_config_var('INCLUDEDIR')+'/qt5',
                                          sysconfig.get_config_var('INCLUDEDIR')+'/<multiarchsubdir>/qt5'],
                            libraries=['Qt5Core', 'Qt5Gui', 'Qt5Widgets']),
                Extension('pybiklib/ext/_qt',
                            ["pybiklib/ext/qt.py"],
                            depends=glob('pybiklib/ext/gl_*.pxd')+[
                                     'pybiklib/ext/qt.pxd', 'pybiklib/ext/qtq.pxd',
                                     'pybiklib/ext/glarea.py', 'pybiklib/ext/qtui.py'],
                            language='c++',
                            include_dirs=[sysconfig.get_config_var('INCLUDEDIR')+'/qt5',
                                          sysconfig.get_config_var('INCLUDEDIR')+'/<multiarchsubdir>/qt5'],
                            libraries=['Qt']),
                Extension('pybiklib/ext/_qtexec',
                            ["pybiklib/ext/qtexec.py"],
                            depends=['pybiklib/ext/qt.pxd'],
                            language='c++',
                            include_dirs=[sysconfig.get_config_var('INCLUDEDIR')+'/qt5',
                                          sysconfig.get_config_var('INCLUDEDIR')+'/<multiarchsubdir>/qt5'],
                            libraries=['Qt5Core', 'Qt5Gui', 'Qt5Widgets']),
            ]
packages = ['pybiklib', 'pybiklib.ext', 'pybiklib.plugins']

cmdclass = {
            'build': build,
            'build_csrc': build_csrc,
            'build_ext': build_ext,
            'build_scripts': build_scripts,
            'build_models': build_models,
            'build_appicons': build_appicons,
            'build_i18n': build_i18n,
            'build_appmeta': build_appmeta,
            'build_man': build_man,
            'build_doc': build_doc,
            'install_data': distutils.command.install_data.install_data,
            'install_lib': install_lib,
            'install_desktop': install_desktop,
            'install': install,
            'clean': clean,
            'sdist': sdist,
        }

setup(
        # Metadata
        name=config.PACKAGE,
        version=config.VERSION,
        author=config.AUTHOR,
        author_email=config.CONTACT_EMAIL,
        url=config.WEBSITE,
        download_url=config.DOWNLOADS,
        
        description=config.SHORT_DESCRIPTION,
        long_description='\n'.join(config.LONG_DESCRIPTION),
        classifiers=[
                'Development Status :: 5 - Production/Stable',
                'Environment :: X11 Applications :: Qt',
                'Intended Audience :: End Users/Desktop',
                'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                'Operating System :: POSIX',
                'Operating System :: POSIX :: Linux',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.4',
                'Programming Language :: Cython',
                #TODO: 'Programming Language :: OpenGL Shading Language',
                'Topic :: Games/Entertainment :: Puzzle Games',
            ],
        keywords='rubik cube puzzle game',
        license=config.LICENSE_NAME,
        # Metadata added in overloaded Distribution
        bug_contact=config.CONTACT_FILEBUG,
        
        # Files
        scripts=scripts,
        data_files=data_files,
        appmeta_templates=appmeta_templates,
        ext_modules=ext_modules,
        packages=packages,
        
        # setup classes
        distclass=Distribution,
        cmdclass=cmdclass,
     )
     



# generated from pybiklib/ext/gl_ogl.pxd

from libc.stddef cimport ptrdiff_t
from libc.stdint cimport int32_t, intptr_t, int8_t, uint8_t


# defines from /usr/include/GL/gl.h:

cdef enum:
    GL_FALSE = 				0
    GL_TRUE = 					1
    GL_UNSIGNED_BYTE = 			0x1401
    GL_FLOAT = 				0x1406
    GL_POINTS = 				0x0000
    GL_LINES = 				0x0001
    GL_TRIANGLES = 				0x0004
    GL_CCW = 					0x0901
    GL_BACK = 					0x0405
    GL_CULL_FACE = 				0x0B44
    GL_DEPTH_TEST = 				0x0B71
    GL_RGB = 					0x1907
    GL_RGBA = 					0x1908
    GL_MAX_TEXTURE_SIZE = 			0x0D33
    GL_TEXTURE_2D = 				0x0DE1
    GL_VENDOR = 				0x1F00
    GL_RENDERER = 				0x1F01
    GL_VERSION = 				0x1F02
    GL_EXTENSIONS = 				0x1F03
    GL_INVALID_ENUM = 				0x0500
    GL_INVALID_VALUE = 			0x0501
    GL_INVALID_OPERATION = 			0x0502
    GL_OUT_OF_MEMORY = 			0x0505
    GL_DEPTH_BUFFER_BIT = 			0x00000100
    GL_COLOR_BUFFER_BIT = 			0x00004000
    GL_MULTISAMPLE = 				0x809D
    GL_SAMPLE_ALPHA_TO_COVERAGE = 		0x809E
    GL_SAMPLE_COVERAGE = 			0x80A0
    GL_SAMPLE_BUFFERS = 			0x80A8
    GL_SAMPLES = 				0x80A9
    GL_SAMPLE_COVERAGE_VALUE = 		0x80AA
    GL_SAMPLE_COVERAGE_INVERT = 		0x80AB


# defines from /usr/include/GL/glext.h:

cdef enum:
    GL_ARRAY_BUFFER =                    0x8892
    GL_STATIC_DRAW =                     0x88E4
    GL_MAX_VERTEX_ATTRIBS =              0x8869
    GL_FRAGMENT_SHADER =                 0x8B30
    GL_VERTEX_SHADER =                   0x8B31
    GL_DELETE_STATUS =                   0x8B80
    GL_COMPILE_STATUS =                  0x8B81
    GL_LINK_STATUS =                     0x8B82
    GL_VALIDATE_STATUS =                 0x8B83
    GL_INFO_LOG_LENGTH =                 0x8B84
    GL_ATTACHED_SHADERS =                0x8B85
    GL_ACTIVE_UNIFORMS =                 0x8B86
    GL_ACTIVE_UNIFORM_MAX_LENGTH =       0x8B87
    GL_ACTIVE_ATTRIBUTES =               0x8B89
    GL_ACTIVE_ATTRIBUTE_MAX_LENGTH =     0x8B8A
    GL_SHADING_LANGUAGE_VERSION =        0x8B8C


# typedefs from /usr/include/GL/gl.h:

IF True:
    ctypedef unsigned int	GLenum
    ctypedef unsigned char	GLboolean
    ctypedef unsigned int	GLbitfield
    ctypedef void		GLvoid
    ctypedef int		GLint
    ctypedef unsigned char	GLubyte
    ctypedef unsigned int	GLuint
    ctypedef int		GLsizei
    ctypedef float		GLfloat
    ctypedef float		GLclampf


# typedefs from /usr/include/GL/glext.h:

IF True:
    ctypedef ptrdiff_t GLsizeiptr
    ctypedef ptrdiff_t GLintptr
    ctypedef char GLchar


# other typedefs:

ctypedef GLchar* const_GLchar_ptr


# functions from /usr/include/GL/gl.h:

IF True:
    cdef void  glClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha ) nogil
    cdef void  glClear( GLbitfield mask ) nogil
    cdef void  glCullFace( GLenum mode ) nogil
    cdef void  glFrontFace( GLenum mode ) nogil
    cdef void  glEnable( GLenum cap ) nogil
    cdef void  glDisable( GLenum cap ) nogil
    cdef GLboolean  glIsEnabled( GLenum cap ) nogil
    cdef void  glGetBooleanv( GLenum pname, GLboolean *params ) nogil
    cdef void  glGetFloatv( GLenum pname, GLfloat *params ) nogil
    cdef void  glGetIntegerv( GLenum pname, GLint *params ) nogil
    cdef GLenum  glGetError() nogil
    cdef GLubyte *  glGetString( GLenum name ) nogil
    cdef void  glViewport( GLint x, GLint y,
                                    GLsizei width, GLsizei height ) nogil
    cdef void  glDrawArrays( GLenum mode, GLint first, GLsizei count ) nogil
    cdef void  glReadPixels( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels ) nogil
    cdef void  glTexImage2D( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLsizei height,
                                    GLint border, GLenum format, GLenum type,
                                    GLvoid *pixels ) nogil
    cdef void  glTexSubImage2D( GLenum target, GLint level,
                                       GLint xoffset, GLint yoffset,
                                       GLsizei width, GLsizei height,
                                       GLenum format, GLenum type,
                                       GLvoid *pixels ) nogil
    cdef void  glActiveTexture( GLenum texture ) nogil


# functions from /usr/include/GL/glext.h:

IF True:
    cdef void  glBindBuffer (GLenum target, GLuint buffer) nogil
    cdef void  glDeleteBuffers (GLsizei n, GLuint *buffers) nogil
    cdef void  glGenBuffers (GLsizei n, GLuint *buffers) nogil
    cdef void  glBufferData (GLenum target, GLsizeiptr size, void *data, GLenum usage) nogil
    cdef void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, void *data) nogil
    cdef void  glAttachShader (GLuint program, GLuint shader) nogil
    cdef void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name) nogil
    cdef void  glCompileShader (GLuint shader) nogil
    cdef GLuint  glCreateProgram () nogil
    cdef GLuint  glCreateShader (GLenum type) nogil
    cdef void  glDeleteProgram (GLuint program) nogil
    cdef void  glDeleteShader (GLuint shader) nogil
    cdef void  glDetachShader (GLuint program, GLuint shader) nogil
    cdef void  glDisableVertexAttribArray (GLuint index) nogil
    cdef void  glEnableVertexAttribArray (GLuint index) nogil
    cdef void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil
    cdef void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name) nogil
    cdef GLint  glGetAttribLocation (GLuint program, GLchar *name) nogil
    cdef void  glGetProgramiv (GLuint program, GLenum pname, GLint *params) nogil
    cdef void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil
    cdef void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params) nogil
    cdef void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog) nogil
    cdef GLint  glGetUniformLocation (GLuint program, GLchar *name) nogil
    cdef void  glLinkProgram (GLuint program) nogil
    cdef void  glShaderSource (GLuint shader, GLsizei count, GLchar **string, GLint *length) nogil
    cdef void  glUseProgram (GLuint program) nogil
    cdef void  glUniform1i (GLint location, GLint v0) nogil
    cdef void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value) nogil
    cdef void  glValidateProgram (GLuint program) nogil
    cdef void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, void *pointer) nogil

# GL version 2.0 needed




#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
from glob import glob
from collections import defaultdict

from . import config
from .ext import qt


class Textures:
    max_size = 256
    stock_dir = os.path.join(config.UI_DIR, 'images')
    
    def __init__(self):
        self.stock_images = {os.path.basename(f): None for f in glob(os.path.join(self.stock_dir, '*'))}
        self.empty_image = (1, 1, b'\x00\x00\x00\x00')
        
    def _get_stock_image(self, name):
        filename = os.path.join(self.stock_dir, name)
        self.stock_images[name] = image = qt.load_image_from_file(filename, self.max_size)
        return image
            
    def get_icons(self):
        return sorted(self.stock_images.keys())
            
    def image_from_file(self, filename):
        if not filename:
            image = self.empty_image
        elif filename.startswith('/'):
            image = qt.load_image_from_file(filename, self.max_size)
        else:
            try:
                image = self.stock_images[filename]
                if image is None:
                    image = self._get_stock_image(filename)
            except KeyError:
                image = None
        return image
        
        
class FaceTheme:
    __slots__ = 'color imagemode image texrect'.split()
    
    
class Theme:
    textures = Textures()
    
    def __init__(self):
        self.faces = defaultdict(FaceTheme)
        
    def load_face_image(self, facekey, filename):
        image = self.textures.image_from_file(filename)
        if image is None:
            self.faces[facekey].image = self.textures.empty_image
            return False
        self.faces[facekey].image = image
        return True
        




# generated with: ./maint.py merge-translators

translators = [
  ('Asturian', 'ast', False, [
    ('Xuacu Saturio', 'https://launchpad.net/~xuacusk8'),
    ('ivarela', 'https://launchpad.net/~ivarela'),
  ]),
  ('Bulgarian', 'bg', False, [
    ('Atanas Kovachki', 'https://launchpad.net/~zdar'),
  ]),
  ('Bosnian', 'bs', False, [
    ('Kenan Dervišević', 'https://launchpad.net/~kenan3008'),
  ]),
  ('Czech', 'cs', False, [
    ('Tadeáš Pařík', 'https://launchpad.net/~pariktadeas'),
  ]),
  ('German', 'de', True, [
    ('B. Clausius', 'https://launchpad.net/~barcc'),
  ]),
  ('Greek', 'el', False, [
    ('George Christofis', 'https://launchpad.net/~geochr'),
    ('mara sdr', 'https://launchpad.net/~paren8esis'),
  ]),
  ('English (United Kingdom)', 'en_GB', True, [
    ('Andi Chandler', 'https://launchpad.net/~bing'),
    ('Anthony Harrington', 'https://launchpad.net/~untaintableangel'),
    ('James Tait', 'https://launchpad.net/~jamestait'),
    ('B. Clausius', 'https://launchpad.net/~barcc'),
  ]),
  ('Spanish', 'es', False, [
    ('Adolfo Jayme', 'https://launchpad.net/~fitojb'),
    ('Paco Molinero', 'https://launchpad.net/~franciscomol'),
    ('Dante Díaz', 'https://launchpad.net/~dante'),
    ('Eduardo Alberto Calvo', 'https://launchpad.net/~edu5800'),
    ('José Lou Chang', 'https://launchpad.net/~obake'),
    ('Oscar Fabian Prieto Gonzalez', 'https://launchpad.net/~ofpprieto'),
    ('Aaron Farias', 'https://launchpad.net/~timido'),
  ]),
  ('Finnish', 'fi', False, [
    ('Jiri Grönroos', 'https://launchpad.net/~jiri-gronroos'),
  ]),
  ('French', 'fr', False, [
    ('Jean-Marc', 'https://launchpad.net/~m-balthazar'),
    ('Nicolas Delvaux', 'https://launchpad.net/~malizor'),
    ('Sylvie Gallet', 'https://launchpad.net/~sylvie-gallet'),
    ('Éfrit', 'https://launchpad.net/~efrit'),
    ('Stanislas Michalak', 'https://launchpad.net/~stanislas-michalak'),
    ('Havok Novak', 'https://launchpad.net/~havok-novak-deactivatedaccount'),
    ('Baptiste Fontaine', 'https://launchpad.net/~bfontaine'),
    ('Célestin Taramarcaz', 'https://launchpad.net/~ctaramarcaz'),
    ('lann', 'https://launchpad.net/~lann'),
    ('Aurélien Ribeiro', 'https://launchpad.net/~aurel-koala'),
    ('Pierre Soulat', 'https://launchpad.net/~pierre-soulat'),
  ]),
  ('Galician', 'gl', True, [
    ('Miguel Anxo Bouzada', 'https://launchpad.net/~mbouzada'),
  ]),
  ('Hebrew', 'he', False, [
    ('Yaron', 'https://launchpad.net/~sh-yaron'),
  ]),
  ('Italian', 'it', False, [
    ('Claudio Arseni', 'https://launchpad.net/~claudio.arseni'),
    ('Gianfranco Frisani', 'https://launchpad.net/~gfrisani'),
    ('acc01291', 'https://launchpad.net/~acc01291'),
  ]),
  ('Kabyle', 'kab', False, [
    ('Belkacem Mohammed', 'https://launchpad.net/~belkacem77'),
  ]),
  ('Malay', 'ms', True, [
    ('abuyop', 'https://launchpad.net/~abuyop'),
  ]),
  ('Polish', 'pl', False, [
    ('Szymon Nieznański', 'https://launchpad.net/~isamu715'),
    ('Michał Rzepiński', 'https://launchpad.net/~micou8'),
  ]),
  ('Brazilian Portuguese', 'pt_BR', False, [
    ('Fábio Nogueira', 'https://launchpad.net/~fnogueira'),
    ('Rodrigo Borges Freitas', 'https://launchpad.net/~rodrigo-borges-freitas'),
    ('Rafael Neri', 'https://launchpad.net/~rafepel'),
  ]),
  ('Russian', 'ru', False, [
    ('Aleksey Kabanov', 'https://launchpad.net/~ak099'),
    ('☠Jay ZDLin☠', 'https://launchpad.net/~black-buddha666'),
    ('Oleg Koptev', 'https://launchpad.net/~koptev-oleg'),
    ('scientistnik', 'https://launchpad.net/~nozdrin-plotnitsky'),
  ]),
  ('Ukrainian', 'uk', True, [
    ('Yuri Chornoivan', 'https://launchpad.net/~yurchor-gmail'),
    ('yurchor', 'https://launchpad.net/~yurchor-deactivatedaccount'),
  ]),
  ('Uzbek', 'uz', False, [
    ('Akmal Xushvaqov', 'https://launchpad.net/~uzadmin'),
  ]),
  ('Chinese (Traditional)', 'zh_TW', False, [
    ('Po-Chun Huang', 'https://launchpad.net/~aphroteus'),
  ]),
]




#!/usr/bin/python3
#-*- coding:utf-8 -*-

#    Copyright © 2009, 2011-2014, 2016-2017  B. Clausius <barcc@gmx.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, re
from contextlib import contextmanager

from .pxmakros import PxMakros
from pybiklib.debug import DEBUG_NOLINENO


class Py2pyxError (Exception): pass

class OpenWriteLazy:
    def __init__(self, path):
        self.path = path
        self.delayed_text = ''
        self.file = None
        
    def write_delayed(self, text):
        self.delayed_text += text
        
    def write(self, text):
        if self.file is None:
            if self.path:
                print('generating', self.path)
                self.file = open(self.path, 'wt', encoding='utf-8')
            else:
                self.file = sys.stdout
        if self.delayed_text:
            self.file.write(self.delayed_text)
            self.delayed_text = ''
        self.file.write(text)
        
    def close(self):
        if self.file is not None:
            self.file.close()
            
            
class PxParser:
    # groups in match objects and patterns
    #   1: leading space
    #   2: px command (includes 3 and 4)
    #   3: pxd flag
    #   4: px key
    #   5: px argument (includes 6 and 7)
    #   6: makro name
    #   7: makro argument
    re_pxline = re.compile(r'^( *)(#px *([a-z]?)(.)|)((\w*)\s*(.*?))\n$', flags=re.ASCII)
    
    # key: px command
    # value:
    #      pattern for pyx-file or None (copy line),
    #      count next lines or "count" (use px argument),
    fdict = {
                '+':    (r'\5',    0),
                '/':    (r'\5',    1),
                '>':    (r'#\2\5',    0),
                '-':    (r'#\2\5',    1),
                ':':    (r'#\2\5',    '*'),
                '.':    (r'#\2\5',    0),
            }
    fdict['#'] = fdict['>']
    
    char_to_makro = {   'd': 'PXDFILE',
                        'h': 'HFILE',
                        'c': 'CFILE',
                        '': '',
                    }
                    
    def __init__(self):
        self.next_cntlines = 0
        
    def parse_line(self, line):
        makro = None
        makrospec = None
        makroarg = None
        match = self.re_pxline.match(line)
        if match.group(2):
            if self.next_cntlines == 0:
                # start px command
                makro = match.group(3)
                if makro == 'm':
                    makro = match.group(6)
                    makrospec = match.expand(r'\2\6')
                    if not makro:
                        raise Py2pyxError('makro name expected')
                    makroarg = match.group(7)
                else:
                    makro = self.char_to_makro[makro]
                    makrospec = match.group(2)
                    makroarg = match.group(5)
            elif self.next_cntlines == 1:
                raise Py2pyxError('#px-command found, normal line expected')
            elif self.next_cntlines == '*':
                # end px command
                if match.group(4) != '.':
                    raise Py2pyxError('closing #px-command or normal line expected')
                if match.group(3):
                    raise Py2pyxError('closing #px-command with makro')
                makro = ''
            else:
                assert False
            pyxpat, self.next_cntlines = self.fdict[match.group(4)]
        elif self.next_cntlines == 0:
            # echo normal line
            makro = ''
            pyxpat = r'\2\5'
            self.next_cntlines = 0
        elif self.next_cntlines == 1:
            # oneline px block
            makroarg = match.group(5)
            pyxpat = r'#\2\5'
            self.next_cntlines = 0
        elif self.next_cntlines == '*':
            # multiline px block
            makroarg = match.group(5)
            pyxpat = r'#\2\5'
        else:
            assert False
        indent = match.group(1)
        pyxline = match.expand(pyxpat)
        
        return indent, makro, makrospec, makroarg, pyxline
        
class PxGenerator (PxMakros):
    def __init__(self, pyxf, pxdf, pxhf, pxcf):
        self.pyxf = pyxf
        self.pxdf = pxdf
        self.pxhf = pxhf
        self.pxcf = pxcf
        self.pyx_indent = ''
        self.pxd_indent = ''
        self.px_skiplines = False
        self.makro = None
        self.makrolineno = None
        self.pyxline = None
        self.namespace = None
        
    def PYXFILE(self, line):
        if self.pyxline is None:
            self.pyxline = line
        elif not line.startswith('#'):
            if self.pyxline.startswith('#'):
                self.pyxline = line
            else:
                raise Py2pyxError('multiple pyx-file lines for source line not allowed: {!r}'.format(line))
                
    def PYXFILE_finalize(self, indent):
        assert self.pyxline is not None
        if not self.pyxline or self.pyxline[0] == '#':
            self.makrospec = ''
        self.pyxf.write(self.pyx_indent + indent + self.pyxline + self.makrospec +'\n')
        self.pyxline = None
            
    def generate_lines(self, lineno, indent, makro, makrospec, makroarg, pyxline):
        if DEBUG_NOLINENO:
            lineno = 'X'
        self.makrospec =  (' '+makrospec) if makrospec else ''
        if self.px_skiplines:
            pyxline = '#~ ' + pyxline
        self.PYXFILE(pyxline)
        
        if makro is not None:
            self.makro = getattr(self, makro, makro)
            self.makrolineno = 0
            self.makrodata = None
        else:
            self.makrolineno += 1
        if self.makro == 'IF':
            bvalue = eval(makroarg)
            assert type(bvalue) is bool
            self.px_skiplines = not bvalue
        elif self.makro == 'IF_END':
            self.px_skiplines = False
        elif self.px_skiplines:
            pass
        elif self.makro == 'INDENT':
            self.pyx_indent = '    '
        elif self.makro == 'DEDENT':
            self.pyx_indent = ''
        elif self.makro == 'DINDENT':
            self.pxd_indent = '    '
        elif self.makro == 'DDEDENT':
            self.pxd_indent = ''
        elif callable(self.makro):
            if self.makro(indent, makroarg, lineno):
                self.makro = None
        elif self.makro:
            raise Py2pyxError('unknown makro: {!s}({!r})'.format(self.makro, makroarg))
                
        self.PYXFILE_finalize(indent)
            
def create_pyx(src_path, pyx_path, pxd_path, pxh_path, pxc_path):
    errors = 0
    openr = lambda path: open(path, 'rt', encoding='utf-8')
    @contextmanager
    def openw(path):
        owl = OpenWriteLazy(path)
        try:
            yield owl
        finally:
            owl.close
            
    with openr(src_path) as srcf, openw(pyx_path) as pyxf, openw(pxd_path) as pxdf, openw(pxh_path) as pxhf, openw(pxc_path) as pxcf:
        pxdf.write_delayed('#generated from: {}\n\n'.format(src_path))
        pxhf.write_delayed('//generated from: {}\n\n'.format(src_path))
        pxcf.write_delayed('//generated from: {}\n\n'.format(src_path))
        pxparser = PxParser()
        pxgenerator = PxGenerator(pyxf, pxdf, pxhf, pxcf)
        for lineno, line in enumerate(srcf):
            try:
                indent, makro, makrospec, makroarg, pyxline = pxparser.parse_line(line)
            except Py2pyxError as e:
                print('%s:%d:'%(src_path, lineno+1), e, file=sys.stderr)
                print('  invalid line:', repr(line), file=sys.stderr)
                indent = ''
                makro = makrospec = makroarg = None
                pyxline = ''
                errors += 1
            except Exception as e:
                print('%s:%d:'%(src_path, lineno+1), e, file=sys.stderr)
                print('  line:', repr(line), file=sys.stderr)
                raise
            try:
                pxgenerator.generate_lines(lineno+1, indent, makro, makrospec, makroarg, pyxline)
            except Py2pyxError as e:
                print('%s:%d:'%(src_path, lineno+1), e, file=sys.stderr)
                print('  invalid line:', repr(line), file=sys.stderr)
                errors += 1
            except Exception as e:
                print('%s:%d:'%(src_path, lineno+1), e, file=sys.stderr)
                print('  line:', repr(line), file=sys.stderr)
                raise
    if errors:
        raise Py2pyxError('create_pyx failed with %s errors' % errors)
        
    
def main(argv):
    if not (1 < len(argv) <= 5):
        print('usage:', os.path.basename(__file__), 'python-file [pyx-filename [pxd-filename [h-filename [c-filename]]]]',
            file=sys.stderr)
        return 1
        
    arg_src = argv[1]
    arg_dst = argv[2] if len(argv) > 2 else None
    arg_pxd = argv[3] if len(argv) > 3 else None
    arg_pxh = argv[4] if len(argv) > 4 else None
    arg_pxc = argv[5] if len(argv) > 5 else None
    if not os.path.exists(arg_src):
        print('error:', arg_src, 'does not exist', file=sys.stderr)
        return 1
    create_pyx(arg_src, arg_dst, arg_pxd, arg_pxh, arg_pxc)
    return 0
    
if __name__ == '__main__':
    sys.exit(main(sys.argv))
    




#-*- coding:utf-8 -*-
# cython: profile=False

#  Copyright © 2009-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# although this file is compiled with Python3 syntax, Cython needs at least division from __future__
from __future__ import print_function, division

# This line makes cython happy
global __name__, __package__    # pylint: disable=W0604
#px/__compiled = True
__compiled = False
#px/DEF OFFSCREEN = False
OFFSCREEN = False

#FIXME: do not use this functions
#px+from libc.stdlib cimport malloc, free
#px+from libc.string cimport memset
#px/from libc.math cimport cos, sin, tan, atan2, sqrt, M_PI, fabs
from math import cos, sin, tan, atan2, sqrt, pi as M_PI, fabs

#pxm>IF '[[GLDEBUG]]' == 'gldebug'
#px+from [[_debug_VARIANT]] cimport *
#pxm>IF '[[GLDEBUG]]' != 'gldebug'
#px+from gl_[[GLVARIANT]] cimport *
#pxm>IF_END
#px+cimport [[_gldraw_VARIANT]] as gldraw

#px:
from . import gldraw
from OpenGL.GLU import * # XXX: without the next line fails with python 3.5
from OpenGL.GL import *
from OpenGL.GL.ARB.vertex_shader import glGetActiveAttribARB as glGetActiveAttrib
#px.

#px/from libc.stdio cimport printf
def printf(fmt, *args): print(fmt % args, end='')
#px/from libc.stdio cimport puts
puts = print

#px/cdef enum: #
if True:
    DEBUG_MSGGL = 0x02
    DEBUG_MSGEXT = 0x08
    DEBUG_DRAW = 0x10
    DEBUG_PICK = 0x20
#px+cdef long debug
debug = 0

def set_debug_flags(module):
    global debug
    if module.DEBUG_MSGGL: debug |= DEBUG_MSGGL
    if module.DEBUG_MSGEXT: debug |= DEBUG_MSGEXT
    if module.DEBUG_DRAW: debug |= DEBUG_DRAW
    if module.DEBUG_PICK: debug |= DEBUG_PICK
    

#px/cdef struct AtlasData:
class AtlasData: pass
    #px+int x, w, h
    #px+char *data
        
#px/cdef struct RenderData:
class renderdata:
    #px+float bg_red
    #px+float bg_green
    #px+float bg_blue
    #px+int y
    #px+int width
    #px+int height
    #px+int pick_x
    #px+int pick_y

    #px+bint multisample
    #px/gldraw.mat4 modelview_matrix
    modelview_matrix = gldraw.mat4()
    #px/gldraw.mat4 projection_matrix
    projection_matrix = gldraw.mat4()
    #px/gldraw.mat4 picking_matrix
    picking_matrix = gldraw.mat4()

    #px+GLuint prog_render
    #px+GLuint projection_location
    #px+GLuint modelview_location
    #px+GLuint prog_hud  # only used if DEBUG_DRAW
    #px+GLuint prog_pick
    #px+GLuint picking_location
    #px+GLuint projection_pick_location
    #px+GLuint modelview_pick_location
    
    #px+bint shader_changed
    #px+char *shader_vertsrc
    #px+char *shader_fragsrc
    #px+char *hud_vertsrc
    #px+char *hud_fragsrc
    #px+char *pick_vertsrc
    #px+char *pick_fragsrc
    
    #px/AtlasData atlasdata[gldraw.MAX_FACES]
    atlasdata = [AtlasData() for i in range(gldraw.MAX_FACES)]
    #px+int atlaslen
    
    #px/float debug_modelview1[3]
    debug_modelview1 = [0.] * 3
    #px/float debug_modelview2[3]
    debug_modelview2 = [0.] * 3
    #px/int debug_viewport1[2]
    debug_viewport1 = [0] * 2
    #px/int debug_viewport2[2]
    debug_viewport2 = [0] * 2
#px+cdef RenderData renderdata


#px/cdef struct UIData:
class uidata:
    #px+bint synced
    
    #px+bint background_changed
    #px+float bg_red
    #px+float bg_green
    #px+float bg_blue
    
    #px+bint viewport_changed
    
    #px+bint rotation_changed
    #px+float rotationxhalf, rotationyhalf
    #px+bint rotation_quat_changed
    #px/gldraw.vec4 rotation_quat
    rotation_quat = gldraw.vec4()
    
    #px+int pick_x
    #px+int pick_y
    
    #px+bint frustum_changed
    #px+float fovy_radius
    #px+float fovy_radius_zoom
    #px+double bounding_sphere_radius
    
    #px+int multisample_changed
    
    #px+char *hud_vertsrc
    #px+char *hud_fragsrc
    #px+char *pick_vertsrc
    #px+char *pick_fragsrc
    
    #px+bint shaders_changed
    #px+char *shader_vertsrc
    #px+char *shader_fragsrc
    
    #px/AtlasData atlasdata[gldraw.MAX_FACES]
    atlasdata = [AtlasData() for i in range(gldraw.MAX_FACES)]
    #px+int atlaslen
    
    #px+bint data_changed
    #px+int vertexdatalen
    #px+char *vertexdata
    #px/long vertexpointers[gldraw.ATTRIB_CNT-1]
    vertexpointers = [None] * (gldraw.ATTRIB_CNT-1)
    #px+int nblocks
    #px/short cnts_block[gldraw.MAX_BLOCKS]
    cnts_block = [None] * gldraw.MAX_BLOCKS
    #px+int idx_debug
    #px+int cnt_debug
    #px+int cnt_pick
    #px+int transformations_count
    #px/gldraw.mat4 transformations[gldraw.MAX_TRANSFORMATIONS]
    transformations = [[[None]*4, [None]*4, [None]*4, [None]*4] for __t in range(gldraw.MAX_TRANSFORMATIONS)]
    
    #px+bint transformation_changed
    #px/short transformation_blocks[gldraw.MAX_BLOCKS]
    transformation_blocks = [None] * gldraw.MAX_BLOCKS
    
    #px+int animation_changed
    #px+int animation_blocks_count
    #px/short animation_blocks[gldraw.MAX_BLOCKS]
    animation_blocks = [None] * gldraw.MAX_BLOCKS
    #px+float angle
    #px+float rotation_x, rotation_y, rotation_z
#px+cdef UIData uidata

class syncpydata: pass
class uipydata: pass
pydata = {}


### quaternion functions

#pxm-FUNC P nogil
def quat_set_identity(quat:'gldraw.vec4 &'):
    quat[0] = 1.
    quat[1] = quat[2] = quat[3] = 0.
    
# (a+bi+cj+dk)(w+xi+yj+zk)
# (aw-bx-cy-dz) + (ax+bw+cz-dy)i + (ay-bz+cw+dx)j + (az+by-cx+dw)k
# ii = jj = kk = ijk = -1

#pxm-FUNC P nogil
def quat_mult(dest:'gldraw.vec4 &', src1:'gldraw.vec4 &', src2:'gldraw.vec4 &'):
    #px+cdef float a,b,c,d
    a = src1[0]*src2[0] - src1[1]*src2[1] - src1[2]*src2[2] - src1[3]*src2[3]
    b = src1[0]*src2[1] + src1[1]*src2[0] + src1[2]*src2[3] - src1[3]*src2[2]
    c = src1[0]*src2[2] - src1[1]*src2[3] + src1[2]*src2[0] + src1[3]*src2[1]
    d = src1[0]*src2[3] + src1[1]*src2[2] - src1[2]*src2[1] + src1[3]*src2[0]
    dest[0] = a
    dest[1] = b
    dest[2] = c
    dest[3] = d
    
##pxm-FUNC P nogil
#def quat_norm(quat:'gldraw.vec4 &'):
#    #px+cdef float n
#    n = sqrt(quat[0]*quat[0] + quat[1]*quat[1] + quat[2]*quat[2] + quat[3]*quat[3])
#    quat[0] /= n
#    quat[1] /= n
#    quat[2] /= n
#    quat[3] /= n
#    if quat[0] < 0.:
#        quat[0] = -quat[0]
#        quat[1] = -quat[1]
#        quat[2] = -quat[2]
#        quat[3] = -quat[3]
        
#pxm-FUNC P nogil
def quat_to_matrix(dest:'gldraw.mat4 &', src:'gldraw.vec4 &'):
    #px+cdef float wx2,wy2,wz2, xx2,xy2,xz2, yy2,yz2,zz2
    #px+cdef float nn
    nn = src[0]*src[0] + src[1]*src[1] + src[2]*src[2] + src[3]*src[3]
    
    wx2 = 2*src[0]*src[1] / nn
    wy2 = 2*src[0]*src[2] / nn
    wz2 = 2*src[0]*src[3] / nn
    xx2 = 2*src[1]*src[1] / nn
    xy2 = 2*src[1]*src[2] / nn
    xz2 = 2*src[1]*src[3] / nn
    yy2 = 2*src[2]*src[2] / nn
    yz2 = 2*src[2]*src[3] / nn
    zz2 = 2*src[3]*src[3] / nn
    
    dest[0][0] = 1-yy2-zz2; dest[1][0] = xy2-wz2;   dest[2][0] = xz2+wy2
    dest[0][1] = xy2+wz2;   dest[1][1] = 1-xx2-zz2; dest[2][1] = yz2-wx2
    dest[0][2] = xz2-wy2;   dest[1][2] = yz2+wx2;   dest[2][2] = 1-xx2-yy2
    
#pxm-FUNC P nogil
def quat_from_angle_axis(quat:'gldraw.vec4 &', angle:float, x:float, y:float, z:float):
    #px+cdef float sa, ca
    ca = cos(angle)
    sa = sin(angle)
    quat[0] = ca
    quat[1] = sa*x
    quat[2] = sa*y
    quat[3] = sa*z
    
### module state

def init_engine():
    gldraw.init_gldraw()
    if debug & DEBUG_MSGEXT:
        print('init module:', __name__)
        print('  from package:', __package__)
        print('  compiled:', __compiled)
        #px+print('  GL-type: [[GLVARIANT]]')
        print('  OFFSCREEN:', OFFSCREEN)
        
    uidata.synced = False
    uidata.background_changed = True
    uidata.viewport_changed = True
    uidata.rotation_changed = True
    uidata.rotation_quat_changed = False
    uidata.frustum_changed = True
    uidata.multisample_changed = 1
    uidata.transformation_changed = False
    uidata.animation_changed = 0
    uidata.data_changed = False
    uidata.shaders_changed = False
    
    uidata.bg_red = 0.0
    uidata.bg_green = 0.0
    uidata.bg_blue = 0.0
    renderdata.y = 0
    renderdata.width = 1
    renderdata.height = 1
    uidata.rotationxhalf = 0.0
    uidata.rotationyhalf = 0.0
    quat_set_identity(uidata.rotation_quat)
    uidata.pick_x = -1
    uidata.pick_y = -1
    renderdata.pick_x = -1
    renderdata.pick_y = -1
    #px+cdef float fovy_angle  # field of view angle
    fovy_angle = 33.0  # field of view angle
    uidata.fovy_radius = tan(fovy_angle * M_PI / 360.0)
    uidata.fovy_radius_zoom = uidata.fovy_radius  # zoom == 1.
    uidata.bounding_sphere_radius = 1.
    uidata.angle = 0
    
    # fill modelview_matrix
    gldraw.matrix_set_identity(renderdata.modelview_matrix)
    # fill projection_matrix
    gldraw.matrix_set_identity(renderdata.projection_matrix)
    renderdata.projection_matrix[2][3] = -1.
    renderdata.projection_matrix[3][3] = 0.
    # fill picking_matrix
    gldraw.matrix_set_identity(renderdata.picking_matrix)
    
    renderdata.shader_changed = False
    renderdata.prog_render = 0
    renderdata.prog_hud = 0
    renderdata.prog_pick = 0
    

#pxm-FUNC P nogil
def _sync_modelview_matrix_translation():
    renderdata.modelview_matrix[3][2] = -uidata.bounding_sphere_radius * (1/uidata.fovy_radius + 1.)
    
#pxm-FUNC P nogil
def _sync_modelview_matrix_rotation():
    #px+cdef gldraw.vec4 *M
    #px+cdef float sx, sy, cx, cy
    #px+cdef float m00, m11, m12, m20
    
    #px/M = <gldraw.vec4*>&renderdata.modelview_matrix[0][0]
    M = renderdata.modelview_matrix
    
    sx = sin(uidata.rotationxhalf)
    sy = sin(uidata.rotationyhalf)
    cx = cos(uidata.rotationxhalf)
    cy = cos(uidata.rotationyhalf)
    
    m00 = 2*cx*cx - 1.
    m11 = 2*cy*cy - 1.
    m12 = 2*sy*cy
    m20 = 2*sx*cx
    # pylint: disable=C0321,C0326
    M[0][0] =  m00;         M[1][0] = 0.;   M[2][0] =  m20
    M[0][1] =  m12 * m20;   M[1][1] = m11;  M[2][1] = -m00 * m12
    M[0][2] = -m11 * m20;   M[1][2] = m12;  M[2][2] =  m00 * m11
    
def set_rotation_rel(x, y):
    #px+cdef gldraw.vec4 quat
    #px+cdef float nxy
    nxy = sqrt(x*x + y*y)
    quat_from_angle_axis(quat, nxy/50., y/nxy, x/nxy, 0.)
    quat_mult(uidata.rotation_quat, quat, uidata.rotation_quat)
    
    # avoid rotation about z-axis
    #uidata.rotation_quat[3] = 0.   # 1.
    #uidata.rotation_quat[3] *= .9  # 2.
    #quat_norm(uidata.rotation_quat)
    # 3.
    #quat_from_angle_axis(quat, fabs(uidata.rotation_quat[3])/10., 0., 0., 1.)
    #quat_mult(uidata.rotation_quat, quat, uidata.rotation_quat)
    
    uidata.rotation_quat_changed = True
    
#pxm-FUNC P nogil
def _sync_projection_matrix():
    # Taken from the OpenGL documentation for glFrustum:
    # glFrustum describes a perspective matrix that produces a perspective projection.
    # The current matrix (see glMatrixMode) is multiplied by this matrix and the result
    # replaces the current matrix, as if glMultMatrix were called with the following
    # matrix as its argument:
    #   [ 2*nearVal                     ]
    #   [ ----------     0        A   0 ]
    #   [ right-left                    ]
    #   [            2*nearVal          ]
    #   [     0      ---------    B   0 ]
    #   [            top-bottom         ]   [p00 0   A   0]
    #   [                               ] = [0   p11 B   0]
    #   [     0          0        C   D ]   [0   0   C   D]
    #   [                               ]   [0   0  -1   0]
    #   [                               ]
    #   [     0          0       -1   0 ]
    #   [                               ]
    #
    #     right+left        top+bottom          farVal+nearVal          2*farVal*nearVal
    # A = ----------    B = ----------    C = - --------------    D = - ----------------
    #     right-left        top-bottom          farVal-nearVal           farVal-nearVal
    #
    # Typically, the matrix mode is GL_PROJECTION, and
    # (left, bottom, -nearVal) and (right, top, -nearVal)
    # specify the points on the near clipping plane that are mapped to the lower left and
    # upper right corners of the window, assuming that the eye is located at (0, 0, 0).
    # -farVal specifies the location of the far clipping plane.
    # Both nearVal and farVal must be positive.
    #
    # right == -left, top == -bottom => A = B = 0
    #
    # p00 = 1 / fovy_radius / aspectx
    # p11 = 1 / fovy_radius / aspecty
    # C = -(1/fovy_radius/zoom + 1.)
    # D = -(1/fovy_radius/zoom + 2.) * bounding_sphere_radius/fovy_radius/zoom
    
    #px+cdef float _1_aspectx, _1_aspecty
    if renderdata.width < renderdata.height:
        _1_aspectx = 1.
        _1_aspecty = renderdata.width / renderdata.height
    else:
        _1_aspectx = renderdata.height / renderdata.width
        _1_aspecty = 1.
    renderdata.projection_matrix[0][0] = _1_aspectx / uidata.fovy_radius_zoom
    renderdata.projection_matrix[1][1] = _1_aspecty / uidata.fovy_radius_zoom
    renderdata.projection_matrix[2][2] = -(1/uidata.fovy_radius + 1.)
    renderdata.projection_matrix[3][2] = -(1/uidata.fovy_radius + 2.
                            ) * uidata.bounding_sphere_radius / uidata.fovy_radius
        
#pxm-FUNC P nogil
def _set_picking_matrix(x:int, y:int):
    # Set picking matrix, restrict drawing to one pixel of the viewport
    # same as:  _glLoadIdentity()
    #           _gluPickMatrix(x, y, 1, 1, viewport)
    # same as:  _glLoadIdentity()
    #           _glTranslatef(renderdata.width - 2*x, renderdata.height - 2*y, 0.)
    #           _glScalef(renderdata.width, renderdata.height, 1.0)
    renderdata.picking_matrix[3][0] = renderdata.width - 2*x
    renderdata.picking_matrix[3][1] = renderdata.height - 2*y
    renderdata.picking_matrix[0][0] = renderdata.width
    renderdata.picking_matrix[1][1] = renderdata.height
    
#pxm-FUNC P nogil
def _set_picking_matrix_identity():
    renderdata.picking_matrix[3][0] = 0.
    renderdata.picking_matrix[3][1] = 0.
    renderdata.picking_matrix[0][0] = 1.
    renderdata.picking_matrix[1][1] = 1.
    
def set_frustum(bounding_sphere_radius, zoom):
    uidata.frustum_changed = True
    if bounding_sphere_radius > 0:
        uidata.bounding_sphere_radius = bounding_sphere_radius
    uidata.fovy_radius_zoom = uidata.fovy_radius / zoom
    
def set_background_color(red, green, blue):
    uidata.background_changed = True
    uidata.bg_red = red
    uidata.bg_green = green
    uidata.bg_blue = blue
    
def set_antialiasing(multisample):
    uidata.multisample_changed = 2 if multisample else 1
    
def set_rotation_xy(x, y):
    x %= 360
    # pylint: disable=C0321
    if y < -120:     y = -120
    elif y > 120:    y = 120
    uidata.rotation_changed = True
    uidata.rotationxhalf = M_PI * x / 360.0
    uidata.rotationyhalf = M_PI * y / 360.0
    return x, y
    
#pxm-FUNC PD nogil
def render_resize(y:int, width:int, height:int):
    renderdata.y = y
    renderdata.width = width
    renderdata.height = height
    uidata.viewport_changed = True
    
def set_transformations(blocks):
    assert uidata.nblocks == len(blocks)
    #px+cdef int i
    #px+cdef short b
    for i, b, unused in blocks:
        uidata.transformation_blocks[i] = b
    uidata.transformation_changed = True
    
#pxm-FUNC P
def set_animation_start(blocks, axisx:float, axisy:float, axisz:float)->'cpdef':
    for i, (b, unused) in enumerate(blocks):
        uidata.animation_blocks[i] = b
    uidata.animation_blocks_count = i+1
    uidata.angle = 0.0
    uidata.rotation_x = axisx
    uidata.rotation_y = axisy
    uidata.rotation_z = axisz
    uidata.animation_changed = 1
    
#pxm-FUNC PD
def set_animation_next(angle:float):
    uidata.angle = -angle
    uidata.animation_changed = 2
    
def set_data(nblocks, vertexdata, vertexpointers, vertexinfo, transformations):
    pydata_postsync()
    #px+cdef int t,i,j
    assert nblocks <= gldraw.MAX_BLOCKS, (nblocks, gldraw.MAX_BLOCKS)
    uidata.nblocks = nblocks
    pydata['vertexdata'] = vertexdata
    uidata.vertexdatalen = len(vertexdata)
    uidata.vertexdata = vertexdata
    assert len(vertexpointers) == gldraw.ATTRIB_CNT-1
    for i in range(gldraw.ATTRIB_CNT-1):
        uidata.vertexpointers[i] = vertexpointers[i]
    cnts_block, uidata.idx_debug, uidata.cnt_debug, uidata.cnt_pick = vertexinfo
    assert len(cnts_block) == nblocks
    for i in range(nblocks):
        uidata.cnts_block[i] = cnts_block[i]
    assert len(transformations) <= gldraw.MAX_TRANSFORMATIONS, len(transformations)
    uidata.transformations_count = len(transformations)
    for t in range(uidata.transformations_count):
        for i in range(4):
            for j in range(4):
                uidata.transformations[t][i][j] = float(transformations[t][i][j])
    uidata.data_changed = True
    
def pydata_postsync():
    global pydata_backed
    if uidata.synced:
        pydata_backed = dict(pydata)
        uidata.synced = False
        
def update_shader_version(shader_src):
    assert shader_src.startswith(b'#version 120\n')
    #px/IF '[[GLVARIANT]]' == 'es2':
    if False:
        shader_src = b'#version 100\n' + shader_src.split(b'\n', 1)[1]
    return shader_src
    
def set_shaders(shader_vertsrc, shader_fragsrc):
    pydata_postsync()
    shader_vertsrc = update_shader_version(shader_vertsrc)
    shader_fragsrc = update_shader_version(shader_fragsrc)
    pydata['shader_vertsrc'] = shader_vertsrc
    uidata.shader_vertsrc = shader_vertsrc
    pydata['shader_fragsrc'] = shader_fragsrc
    uidata.shader_fragsrc = shader_fragsrc
    uidata.shaders_changed = True
    
def set_fixedshaders(fixedshaders):
    pydata_postsync()
    fixedshaders = [update_shader_version(s) for s in fixedshaders]
    pydata['fixedshaders'] = fixedshaders
    uidata.pick_vertsrc = fixedshaders[0]
    uidata.pick_fragsrc = fixedshaders[1]
    if debug & DEBUG_DRAW:
        uidata.hud_vertsrc = fixedshaders[2]
        uidata.hud_fragsrc = fixedshaders[1]
    
#pxm-FUNC PD nogil
def sync_set_fixedshaders():
    renderdata.pick_vertsrc = uidata.pick_vertsrc
    renderdata.pick_fragsrc = uidata.pick_fragsrc
    if debug & DEBUG_DRAW:
        renderdata.hud_vertsrc = uidata.hud_vertsrc
        renderdata.hud_fragsrc = uidata.hud_fragsrc
    
def set_pick_position(x, y):
    uidata.pick_x = x
    uidata.pick_y = y
    
#pxm-FUNC PD nogil
def sync_set_pick_position():
    renderdata.pick_x = uidata.pick_x
    renderdata.pick_y = uidata.pick_y
    
#pxm-FUNC P
def set_atlas_data(i:int, data:bytes, x:int, w:int, h:int)->'cpdef':
    assert i < gldraw.MAX_FACES, (i, gldraw.MAX_FACES)
    if i == 0:
        uipydata.atlasdata = []
    uipydata.atlasdata.append(data)
    uidata.atlasdata[i].x = x
    uidata.atlasdata[i].w = w
    uidata.atlasdata[i].h = h
    uidata.atlasdata[i].data = data
    uidata.atlaslen = i+1
    
#pxm-FUNC PD nogil
def sync_set_atlas_data():
    #px+cdef int i
    for i in range(uidata.atlaslen):
        renderdata.atlasdata[i].x = uidata.atlasdata[i].x
        renderdata.atlasdata[i].w = uidata.atlasdata[i].w
        renderdata.atlasdata[i].h = uidata.atlasdata[i].h
        renderdata.atlasdata[i].data = uidata.atlasdata[i].data
    renderdata.atlaslen = uidata.atlaslen
    
#pxm-FUNC PD nogil
def sync():
    uidata.synced = True
    if uidata.background_changed:
        renderdata.bg_red = uidata.bg_red
        renderdata.bg_green = uidata.bg_green
        renderdata.bg_blue = uidata.bg_blue
        uidata.background_changed = False
    if uidata.viewport_changed:
        _sync_projection_matrix()
        uidata.viewport_changed = False
    if uidata.rotation_changed:
        _sync_modelview_matrix_rotation()
        uidata.rotation_changed = False
    if uidata.rotation_quat_changed:
        quat_to_matrix(renderdata.modelview_matrix, uidata.rotation_quat)
        uidata.rotation_quat_changed = False
    if uidata.frustum_changed:
        _sync_modelview_matrix_translation()
        _sync_projection_matrix()
        uidata.frustum_changed = False
    if uidata.multisample_changed:
        renderdata.multisample = uidata.multisample_changed - 1
        uidata.multisample_changed = 0
    if uidata.data_changed:
        gldraw.sync_vertexdata(uidata.vertexdatalen, uidata.vertexdata, uidata.vertexpointers)
        gldraw.sync_blocks(uidata.nblocks, uidata.cnts_block, uidata.idx_debug, uidata.cnt_debug, uidata.cnt_pick)
        gldraw.sync_transformations(uidata.transformations_count, uidata.transformations)
        uidata.data_changed = False
    if uidata.transformation_changed:
        gldraw.sync_block_transformations(uidata.transformation_blocks)
        uidata.transformation_changed = False
    if uidata.animation_changed == 2:
        gldraw.sync_animation_next(uidata.angle, uidata.rotation_x, uidata.rotation_y, uidata.rotation_z)
        uidata.animation_changed = 0
    elif uidata.animation_changed == 1:
        gldraw.sync_animation_start(uidata.animation_blocks_count, uidata.animation_blocks)
        uidata.animation_changed = 0
    if uidata.shaders_changed:
        renderdata.shader_changed = True
        renderdata.shader_vertsrc = uidata.shader_vertsrc
        renderdata.shader_fragsrc = uidata.shader_fragsrc
        uidata.shaders_changed = False
        
### GL state

#pxm-FUNC P nogil
def _gl_print_string(msg:'char *', name:'GLenum'):
    #px/printf('%s %s\n', msg, <char*>glGetString(name))
    printf('%s %s\n', msg, glGetString(name))
    
#pxm-FUNC P nogil
def _gl_print_float(msg:'char *', name:'GLenum'):
    #px+cdef GLfloat i
    #px/glGetFloatv(name, &i)
    i = glGetFloatv(name)
    printf('%s %f\n', msg, i)
    
#pxm-FUNC P nogil
def _gl_print_integer(msg:'char *', name:'GLenum'):
    #px+cdef GLint i
    #px/glGetIntegerv(name, &i)
    i = glGetIntegerv(name)
    printf('%s %d\n', msg, i)
    
#pxm-FUNC P nogil
def _gl_print_bool(msg:'char *', name:'GLenum'):
    #px+cdef GLboolean i
    #px+glGetBooleanv(name, &i)
    #px/printf('%s %d\n', msg, <int>i)
    printf('%s %d\n', msg, glGetBooleanv(name))
    
#pxm-FUNC PD nogil
def gl_init():
    if debug & DEBUG_MSGGL:
        puts('GL Strings:')
        _gl_print_string('  GL Vendor:', GL_VENDOR)
        _gl_print_string('  GL Renderer:', GL_RENDERER)
        _gl_print_string('  GL Version:', GL_VERSION)
        _gl_print_string('  GL Shading Language Version:', GL_SHADING_LANGUAGE_VERSION)
        #_gl_print_string('  GL Extensions:', GL_EXTENSIONS)
        _gl_print_integer('  GL_SAMPLE_BUFFERS:', GL_SAMPLE_BUFFERS)
        _gl_print_float('  GL_SAMPLE_COVERAGE_VALUE:', GL_SAMPLE_COVERAGE_VALUE)
        _gl_print_bool('  GL_SAMPLE_COVERAGE_INVERT:', GL_SAMPLE_COVERAGE_INVERT)
        _gl_print_integer('  GL_SAMPLES:', GL_SAMPLES)
        #px/IF '[[GLVARIANT]]' == 'ogl':
        if True:
            printf('  GL_MULTISAMPLE: %d\n', glIsEnabled(GL_MULTISAMPLE))
        #print('  GL_SAMPLE_ALPHA_TO_COVERAGE:', glIsEnabled(GL_SAMPLE_ALPHA_TO_COVERAGE))
        #print('  GL_SAMPLE_COVERAGE:', glIsEnabled(GL_SAMPLE_COVERAGE))
        _gl_print_integer('  GL_MAX_VERTEX_ATTRIBS:', GL_MAX_VERTEX_ATTRIBS)
        _gl_print_integer('  GL_MAX_TEXTURE_SIZE:', GL_MAX_TEXTURE_SIZE)
    gldraw.gl_init_buffers()
    if debug & DEBUG_DRAW:
        gl_create_hud_program()
    gl_create_pick_program()
    
#pxm-FUNC PD nogil
def gl_exit():
    gldraw.gl_delete_buffers()
    if renderdata.prog_render > 0:  glDeleteProgram(renderdata.prog_render)
    if renderdata.prog_hud > 0:     glDeleteProgram(renderdata.prog_hud)
    if renderdata.prog_pick > 0:    glDeleteProgram(renderdata.prog_pick)
    renderdata.prog_render = 0
    renderdata.prog_hud = 0
    renderdata.prog_pick = 0
    
### render functions

#pxm-FUNC PD nogil
def gl_set_atlas_texture(width:int, height:int):
    #px+cdef int i, x, w, h
    #px+cdef char *data = <char*>malloc(width*height*4)
    #px+memset(data, 0, width*height*4)
    #FIXME: GL_INVALID_OPERATION, but it works, at least on the phone a call to glTexImage2D is needed before glTexSubImage2D
    #px+glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data)
    #px+free(data)
    for i in range(renderdata.atlaslen):
        x, w, h = renderdata.atlasdata[i].x, renderdata.atlasdata[i].w, renderdata.atlasdata[i].h
        data = renderdata.atlasdata[i].data
        glTexSubImage2D(GL_TEXTURE_2D, 0, x, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE, data)
    
#pxm-FUNC P nogil
def _gl_set_matrix(location:'GLint', matrix:'gldraw.mat4 &'):
    #px/glUniformMatrix4fv(location, 1, GL_FALSE, &matrix[0][0])
    glUniformMatrix4fv(location, 1, GL_FALSE, matrix)
    
#pxm-FUNC PD nogil
def gl_render():
    if renderdata.shader_changed:
        gl_create_render_program()
        renderdata.shader_changed = False
    glEnable(GL_DEPTH_TEST)
    glEnable(GL_CULL_FACE)
    glCullFace(GL_BACK)
    glFrontFace(GL_CCW)
    gldraw.gl_enable_data()
    glViewport(0, renderdata.y, renderdata.width, renderdata.height)
    if debug & DEBUG_PICK:
        _set_picking_matrix_identity()
        _gl_render_pick()
    else:
        glUseProgram(renderdata.prog_render)
        #px/IF '[[GLVARIANT]]' == 'ogl':
        if True:
            if renderdata.multisample:
                glEnable(GL_MULTISAMPLE)
            else:
                glDisable(GL_MULTISAMPLE)
        #px+IF OFFSCREEN:
        #px+    glClearColor(renderdata.bg_red, renderdata.bg_green, renderdata.bg_blue, 0.)
        #px/ELSE:
        if True:
            glClearColor(renderdata.bg_red, renderdata.bg_green, renderdata.bg_blue, 1.)
        _gl_set_matrix(renderdata.projection_location, renderdata.projection_matrix)
        _gl_set_matrix(renderdata.modelview_location, renderdata.modelview_matrix)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        gldraw.gl_draw_cube()
    if debug & DEBUG_DRAW:
        gldraw.gl_draw_cube_debug()
    glClearColor(0, 0, 0, 0)
    glUseProgram(0)
    gldraw.gl_disable_data()
    glDisable(GL_CULL_FACE)
    glDisable(GL_DEPTH_TEST)
        
#pxm-FUNC PD nogil
def gl_render_select_debug():
    #px/cdef GLfloat selectdata[12]
    selectdata = [0.] * 12
    selectdata[0] = renderdata.debug_modelview1[0]
    selectdata[1] = renderdata.debug_modelview1[1]
    selectdata[2] = renderdata.debug_modelview1[2]
    selectdata[3] = renderdata.debug_modelview2[0]
    selectdata[4] = renderdata.debug_modelview2[1]
    selectdata[5] = renderdata.debug_modelview2[2]
    selectdata[6] = renderdata.debug_viewport1[0] / renderdata.width * 2 - 1
    selectdata[7] = renderdata.debug_viewport1[1] / renderdata.height * 2 - 1
    selectdata[8] = 0
    selectdata[9] = renderdata.debug_viewport2[0] / renderdata.width * 2 - 1
    selectdata[10] = renderdata.debug_viewport2[1] / renderdata.height * 2 - 1
    selectdata[11] = 0
    gldraw.gl_enable_data()
    glUseProgram(renderdata.prog_render)
    glViewport(0, renderdata.y, renderdata.width, renderdata.height)
    #px/gldraw.gl_draw_select_debug(&selectdata[0], sizeof(selectdata), renderdata.prog_hud)
    gldraw.gl_draw_select_debug(selectdata, 0, renderdata.prog_hud)
    glUseProgram(0)
    gldraw.gl_disable_data()
    

### picking functions

#pxm-FUNC P nogil
def _gl_render_pick():
    glUseProgram(renderdata.prog_pick)
    #px/IF '[[GLVARIANT]]' == 'ogl':
    if True:
        glDisable(GL_MULTISAMPLE)
    glClearColor(0., 0., 0., 1.)
    _gl_set_matrix(renderdata.picking_location, renderdata.picking_matrix)
    _gl_set_matrix(renderdata.projection_pick_location, renderdata.projection_matrix)
    _gl_set_matrix(renderdata.modelview_pick_location, renderdata.modelview_matrix)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    gldraw.gl_pick_cube()
    
#pxm-FUNC PD nogil
def gl_pick_polygons()->int:
    #px+cdef unsigned char pixel[4]
    #px+cdef int index
    
    if not (0 <= renderdata.pick_x < renderdata.width and 0 <= renderdata.pick_y < renderdata.height):
        return 0
    _set_picking_matrix(renderdata.pick_x, renderdata.pick_y)
    glEnable(GL_DEPTH_TEST)
    glEnable(GL_CULL_FACE)
    glCullFace(GL_BACK)
    glFrontFace(GL_CCW)
    gldraw.gl_enable_data()
    glViewport(0, 0, 1, 1)
    _gl_render_pick()
    gldraw.gl_disable_data()
    glDisable(GL_CULL_FACE)
    glDisable(GL_DEPTH_TEST)
    #px/glReadPixels(0, 0, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, pixel)
    pixel = glReadPixels(0, 0, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, [[[0, 0, 0, 0]]])[0][0]
    return pixel[0]<<4 | pixel[1] | pixel[2]>>4
    
#pxm-FUNC P nogil
def _modelview_to_viewport(vvect:'float *', mvect:'int *'):
    #px+cdef gldraw.vec4 *M
    #px+cdef gldraw.vec4 *P
    #px+cdef float u0, u1, u2, v0, v1, v3
    
    #px/M = <gldraw.vec4*>&renderdata.modelview_matrix[0][0]
    M = renderdata.modelview_matrix
    #px/P = <gldraw.vec4*>&renderdata.projection_matrix[0][0]
    P = renderdata.projection_matrix
    
    # u = M^T * vvect
    #assert M[1][0] == 0
    u0 = M[0][0]*vvect[0] + M[1][0]*vvect[1] + M[2][0]*vvect[2] + M[3][0]
    u1 = M[0][1]*vvect[0] + M[1][1]*vvect[1] + M[2][1]*vvect[2] + M[3][1]
    u2 = M[0][2]*vvect[0] + M[1][2]*vvect[1] + M[2][2]*vvect[2] + M[3][2]
    #u3 = 1.
    
    # v = P * u
    v0 = P[0][0] * u0 + P[1][0] * u1 + P[2][0] * u2 + P[3][0] #* u3
    v1 = P[0][1] * u0 + P[1][1] * u1 + P[2][1] * u2 + P[3][1] #* u3
    #v2 = P[0][2] * u0 + P[1][2] * u1 + P[2][2] * u2 + P[3][2] * u3
    v3 = P[0][3] * u0 + P[1][3] * u1 + P[2][3] * u2 + P[3][3] #* u3
    
    mvect[0] = int((v0 / v3 + 1) / 2 * renderdata.width)
    mvect[1] = int((v1 / v3 + 1) / 2 * renderdata.height)
    
#pxm-FUNC P nogil
def _viewport_to_modelview(vvect:'int *', mvect:'float *'):
    #px+cdef gldraw.vec4 *MT
    #px+cdef gldraw.vec4 *P
    #px+cdef float u0, u1, u2, u3, v0, v1, v2
    
    v0 = vvect[0] / renderdata.width * 2 - 1
    v1 = vvect[1] / renderdata.height * 2 - 1
    v2 = 0
    #v3 = 1
    
    # P * P^-1 = I
    # a 0 0 0     A 0 0 0     aA 0  0  0      E = -eD/c
    # 0 b 0 0  *  0 B 0 0  =  0  bB 0  0 ,    
    # 0 0 e c     0 0 0 D     0  0  cC 0      
    # 0 0 d 0     0 0 C E     0  0  0  dD     
    #px/P = <gldraw.vec4*>&renderdata.projection_matrix[0][0]
    P = renderdata.projection_matrix
    # u = P^-1 * v
    u0 = 1 / P[0][0] * v0
    u1 = 1 / P[1][1] * v1
    u2 = 1 / P[2][3]
    #assert u2 == -1
    u3 = - P[2][2] / P[3][2] / P[2][3]
    
    # MT * MT^-1 = I
    # a b c 0     a 0 f 0     1 0 0 0    U = -cz
    # 0 d e 0  *  b d g 0  =  0 1 0 0,   V = -ez
    # f g h 0     c e h 0     0 0 1 0    W = -hz
    # 0 0 z 1     U V W 1     0 0 0 1
    #px/MT = <gldraw.vec4*>&renderdata.modelview_matrix[0][0]
    MT = renderdata.modelview_matrix
    # v = M^-1 * u
    v0 = MT[0][0]*u0 + MT[0][1]*u1 + MT[0][2]*u2 - MT[0][2] * MT[3][2]*u3
    v1 = MT[1][0]*u0 + MT[1][1]*u1 + MT[1][2]*u2 - MT[1][2] * MT[3][2]*u3
    v2 = MT[2][0]*u0 + MT[2][1]*u1 + MT[2][2]*u2 - MT[2][2] * MT[3][2]*u3
    #v3 = u3
    
    mvect[0] = v0 / u3
    mvect[1] = v1 / u3
    mvect[2] = v2 / u3
    
def get_cursor_angle(d):
    #px+cdef float angle
    #px+cdef int i
    #px+cdef float x, y
    renderdata.debug_viewport1[0] = uidata.pick_x
    renderdata.debug_viewport1[1] = uidata.pick_y
    _viewport_to_modelview(renderdata.debug_viewport1, renderdata.debug_modelview1)
    for i in range(3):
        renderdata.debug_modelview2[i] = d[i] + renderdata.debug_modelview1[i]
    _modelview_to_viewport(renderdata.debug_modelview2, renderdata.debug_viewport2)
    x = renderdata.debug_viewport1[0] - renderdata.debug_viewport2[0]
    y = renderdata.debug_viewport1[1] - renderdata.debug_viewport2[1]
    angle = atan2(x, y) * 180.0 / M_PI
    return angle
    
    
### shader functions

#pxm-FUNC P nogil
def _gl_print_shader_log(shader:'GLuint'):
    #px+cdef GLint log_len
    #px+cdef GLsizei length
    #px+cdef char log[1024]
    #px/glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &log_len)
    log_len = glGetShaderiv(shader, GL_INFO_LOG_LENGTH)
    if log_len > 0:
        puts('==== shader info log:')
        #px/glGetShaderInfoLog(shader, 1023, &length, log)
        log = glGetShaderInfoLog(shader).decode('utf-8')
        printf('%s', log)
        puts('====')
    else:
        puts('==== empty shader info log')
        
#pxm-FUNC P nogil
def _gl_print_program_log(program:'GLuint'):
    #px+cdef GLint log_len
    #px+cdef GLsizei length
    #px+cdef char log[1024]
    #px/glGetProgramiv(program, GL_INFO_LOG_LENGTH, &log_len)
    log_len = glGetProgramiv(program, GL_INFO_LOG_LENGTH)
    if log_len > 0:
        puts('==== program info log:')
        #px/glGetProgramInfoLog(program, 1023, &length, log)
        log = glGetProgramInfoLog(program).decode('utf-8')
        printf('%s', log)
        puts('====')
    else:
        puts('==== empty program info log')
        
#pxm-FUNC P nogil
def _gl_create_compiled_shader(shadertype:'GLenum', source:'char *')->'GLuint':
    #px+cdef GLuint shader
    #px+cdef const_GLchar_ptr pchar
    #px+cdef GLint compile_status
    
    shader = glCreateShader(shadertype)
    if shader == 0:
        puts('Failed to create shader')
        return 0
    #px+pchar = <const_GLchar_ptr><char*>source
    #px/glShaderSource(shader, 1, &pchar, NULL)
    glShaderSource(shader, source.decode('utf-8')) # PyOpenGL needs str currently
    glCompileShader(shader)
    if debug & DEBUG_MSGGL:
        #px/glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &compile_status)
        compile_status = glGetShaderiv(shader, GL_INFO_LOG_LENGTH)
        printf('GL_INFO_LOG_LENGTH %d\n', compile_status)
        compile_status = compile_status <= 1
        if compile_status:
            #px/glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_status)
            compile_status = glGetShaderiv(shader, GL_COMPILE_STATUS)
    else:
        #px/glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_status)
        compile_status = glGetShaderiv(shader, GL_COMPILE_STATUS)
    if not compile_status:
        _gl_print_shader_log(shader)
    return shader
    
#pxm-FUNC P nogil
def _get_program_iv(program:'GLuint', pname:'GLenum')->'GLint':
    #px+cdef GLint param
    #px/glGetProgramiv(program, pname, &param)
    param = int(glGetProgramiv(program, pname))
    return param
    
#pxm-FUNC P nogil
def _gl_print_program_info(program:'GLuint')->'GLint':
    #px+cdef GLint link_status, info_log_length, param, aaml, auml
    #px+cdef int i
    #px+cdef GLsizei alength
    #px+cdef GLint asize, location
    #px+cdef GLenum atype
    #px+cdef char aname[1024]
    
    printf('shader program info %d\n', program)
    glValidateProgram(program)
    printf('  delete status %d\n',   _get_program_iv(program, GL_DELETE_STATUS))
    link_status = _get_program_iv(program, GL_LINK_STATUS)
    printf('  link status %d\n',     link_status)
    printf('  validate status %d\n', _get_program_iv(program, GL_VALIDATE_STATUS))
    info_log_length = _get_program_iv(program, GL_INFO_LOG_LENGTH)
    printf('  info log length %d\n', info_log_length)
    printf('  attached shaders %d\n',_get_program_iv(program, GL_ATTACHED_SHADERS))
    aaml = _get_program_iv(program, GL_ACTIVE_ATTRIBUTE_MAX_LENGTH)
    printf('  active attribute max length %d\n', aaml)
    auml = _get_program_iv(program, GL_ACTIVE_UNIFORM_MAX_LENGTH)
    printf('  active uniform max length %d\n', auml)
    
    #px/glGetProgramiv(program, GL_ACTIVE_ATTRIBUTES, &param)
    param = glGetProgramiv(program, GL_ACTIVE_ATTRIBUTES)
    printf('active attributes %d:\n', param)
    for i in range(param):
        #px/glGetActiveAttrib(program, i, 1023, &alength, &asize, &atype, aname)
        aname, asize, atype = glGetActiveAttrib(program, i); alength = len(aname)
        location = glGetAttribLocation(program, aname)
        printf('  %d %-*s length=%d size=%d type=%d location=%d\n', i, aaml+3, aname, alength, asize, atype, location)
        
    #px/glGetProgramiv(program, GL_ACTIVE_UNIFORMS, &param)
    param = glGetProgramiv(program, GL_ACTIVE_UNIFORMS)
    printf('active uniforms %d:\n', param)
    for i in range(param):
        #px/glGetActiveUniform(program, i, 1023, &alength, &asize, &atype, aname)
        aname, asize, atype = glGetActiveUniform(program, i); alength = len(aname)
        location = glGetUniformLocation(program, aname)
        printf('  %d %-*s length=%d size=%d type=%d location=%d\n', i, auml+3, aname, alength, asize, atype, location)
    return link_status and info_log_length <= 1
        
#pxm-FUNC P nogil
def _gl_program_add_shaders(program:'GLuint', vertex_source:'char *', fragment_source:'char *'):
    #px+cdef GLuint vertex_shader = 0, fragment_shader = 0
    #px+cdef GLint link_status
    
    if debug & DEBUG_MSGGL: puts('  creating vertex shader')
    vertex_shader = _gl_create_compiled_shader(GL_VERTEX_SHADER, vertex_source)
    if vertex_shader:
        glAttachShader(program, vertex_shader)
    else:
        puts('error: no vertex shader')
        
    if debug & DEBUG_MSGGL: puts('  creating fragment shader')
    fragment_shader = _gl_create_compiled_shader(GL_FRAGMENT_SHADER, fragment_source)
    if fragment_shader:
        glAttachShader(program, fragment_shader)
    else:
        puts('error: no fragment shader')
        
    glLinkProgram(program)
    if debug & DEBUG_MSGGL:
        link_status = _gl_print_program_info(program)
    else:
        #px/glGetProgramiv(program, GL_LINK_STATUS, &link_status)
        link_status = glGetProgramiv(program, GL_LINK_STATUS)
    if not link_status:
        printf('link status %d\n', link_status)
        _gl_print_program_log(program)
    if vertex_shader:
        glDetachShader(program, vertex_shader)
        glDeleteShader(vertex_shader)
    if fragment_shader:
        glDetachShader(program, fragment_shader)
        glDeleteShader(fragment_shader)
    
#pxm-FUNC PD nogil
def gl_create_render_program():
    #px+cdef GLint location
    if debug & DEBUG_MSGGL:  puts('Creating render shaders:')
    if renderdata.prog_render > 0:
        glDeleteProgram(renderdata.prog_render)
    renderdata.prog_render = glCreateProgram()
    glBindAttribLocation(renderdata.prog_render, gldraw.ATTRIB_LOCATION+0, b'vertex_attr')
    glBindAttribLocation(renderdata.prog_render, gldraw.ATTRIB_LOCATION+1, b'normal_attr')
    glBindAttribLocation(renderdata.prog_render, gldraw.ATTRIB_LOCATION+2, b'color_attr')
    glBindAttribLocation(renderdata.prog_render, gldraw.ATTRIB_LOCATION+3, b'texcoord_attr')
    glBindAttribLocation(renderdata.prog_render, gldraw.ATTRIB_LOCATION+4, b'barycentric_attr')
    _gl_program_add_shaders(renderdata.prog_render, renderdata.shader_vertsrc, renderdata.shader_fragsrc)
    if renderdata.prog_render > 0:
        glUseProgram(renderdata.prog_render)
        location = glGetUniformLocation(renderdata.prog_render, b'tex')
        glUniform1i(location, 0) # 0 is the texture unit (-> glActiveTexture)
        renderdata.modelview_location = glGetUniformLocation(renderdata.prog_render, b'modelview')
        renderdata.projection_location = glGetUniformLocation(renderdata.prog_render, b'projection')
        location = glGetUniformLocation(renderdata.prog_render, b'object')
        gldraw.gl_init_object_location(location)
        
#pxm-FUNC P nogil
def gl_create_hud_program():
    #px+cdef GLint location
    if debug & DEBUG_MSGGL: puts('Creating "hud" shaders:')
    if renderdata.prog_hud > 0:
        glDeleteProgram(renderdata.prog_hud)
    renderdata.prog_hud = glCreateProgram()
    glBindAttribLocation(renderdata.prog_hud, gldraw.ATTRIB_LOCATION+0, b'vertex_attr')
    glBindAttribLocation(renderdata.prog_hud, gldraw.ATTRIB_LOCATION+2, b'color_attr')
    _gl_program_add_shaders(renderdata.prog_hud, renderdata.hud_vertsrc, renderdata.hud_fragsrc)
    if renderdata.prog_hud > 0:
        glUseProgram(renderdata.prog_hud)
    
#pxm-FUNC P nogil
def gl_create_pick_program():
    if debug & DEBUG_MSGGL: puts('Creating "pick" shaders:')
    if renderdata.prog_pick > 0:
        glDeleteProgram(renderdata.prog_pick)
    renderdata.prog_pick = glCreateProgram()
    glBindAttribLocation(renderdata.prog_pick, gldraw.PICKATTRIB_LOCATION+0, b'vertex_attr')
    glBindAttribLocation(renderdata.prog_pick, gldraw.PICKATTRIB_LOCATION+1, b'color_attr')
    _gl_program_add_shaders(renderdata.prog_pick, renderdata.pick_vertsrc, renderdata.pick_fragsrc)
    if renderdata.prog_pick > 0:
        glUseProgram(renderdata.prog_pick)
        renderdata.picking_location = glGetUniformLocation(renderdata.prog_pick, b'picking')
        renderdata.projection_pick_location = glGetUniformLocation(renderdata.prog_pick, b'projection')
        renderdata.modelview_pick_location = glGetUniformLocation(renderdata.prog_pick, b'modelview')
    




# -*- coding: utf-8 -*-

#  Copyright © 2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


def invert(game):
    game.move_sequence.invert()
    game.recalc_current_state()
    
def normalize_complete_rotations(game):
    game.move_sequence.normalize_complete_rotations(game.initial_state.model)
    game.recalc_current_state()
    
def normalize_moves(game):
    game.move_sequence.normalize_moves(game.initial_state.model)
    game.recalc_current_state()
    

